import random

import _mjx

import mjx.agent
import mjx.env


def test_EnvRunner():
    agents = {
        "player_0": _mjx.RandomDebugAgent(),
        "player_1": _mjx.RandomDebugAgent(),
        "player_2": _mjx.RandomDebugAgent(),
        "player_3": _mjx.RandomDebugAgent(),
    }
    seed_generator = _mjx.RandomSeedGenerator(["player_0", "player_1", "player_2", "player_3"])
    runner = _mjx.EnvRunner(agents, seed_generator, 100, 4, 10, None, None)


def test_RandomSeedGenerator():
    seed_generator = _mjx.RandomSeedGenerator(["player_0", "player_1", "player_2", "player_3"])  # type: ignore
    N = 100000
    seeds = set([])
    first_dealer_cnt = {}
    for i in range(N):
        seed, player_ids = seed_generator.get()
        seeds.add(seed)
        first_dealer_cnt[player_ids[0]] = (
            first_dealer_cnt[player_ids[0]] + 1 if player_ids[0] in first_dealer_cnt else 1
        )

    assert len(seeds) == N
    assert N / 4 - N / 10 < first_dealer_cnt["player_0"] < N / 4 + N / 10
    assert N / 4 - N / 10 < first_dealer_cnt["player_1"] < N / 4 + N / 10
    assert N / 4 - N / 10 < first_dealer_cnt["player_2"] < N / 4 + N / 10
    assert N / 4 - N / 10 < first_dealer_cnt["player_3"] < N / 4 + N / 10


def test_DuplicateRandomSeedGenerator():
    seed_generator = _mjx.DuplicateRandomSeedGenerator(["player_0", "player_1", "player_2", "player_3"])  # type: ignore
    N = 100000
    seeds = set([])
    first_dealer_cnt = {}
    for i in range(N):
        seed, player_ids = seed_generator.get()
        seeds.add(seed)
        first_dealer_cnt[player_ids[0]] = (
            first_dealer_cnt[player_ids[0]] + 1 if player_ids[0] in first_dealer_cnt else 1
        )

    assert len(seeds) == N // 4
    assert N // 4 == first_dealer_cnt["player_0"]
    assert N // 4 == first_dealer_cnt["player_1"]
    assert N // 4 == first_dealer_cnt["player_2"]
    assert N // 4 == first_dealer_cnt["player_3"]


def test_MjxEnv():
    random_agent = mjx.agent.RandomDebugAgent()

    random.seed(1234)
    env = mjx.env.MjxEnv()
    obs_dict = env.reset(1234)

    assert len(obs_dict) == 1
    assert "player_2" in obs_dict
    assert env.state().to_proto().hidden_state.wall[:5] == [24, 3, 87, 124, 97]

    while not env.done():
        action_dict = {}
        for agent, obs in obs_dict.items():
            action_dict[agent] = random_agent.act(obs)
        obs_dict = env.step(action_dict)
        if not env.done():
            rewards = env.rewards()
            assert len(rewards) == 4
            assert rewards["player_0"] == 0
            assert rewards["player_1"] == 0
            assert rewards["player_2"] == 0
            assert rewards["player_3"] == 0

    # MjxEnvのDone周りの挙動についての仕様は
    # https://github.com/mjx-project/mjx/pull/1055
    # を参考
    # 基本的な考えとしては
    #  - Dummyアクションは情報共有のためにあるので、Dummyを含む
    #    Obsが返ってくるタイミングで、doneがtrue, non-zero rewardsが得られる
    #  - Dummyアクションを取ると、次の局に進む（オーラス）は空obsが返ってくる
    #  - ユーザ側から見れば、doneがtrueのタイミングでobsやrewardsにアクセスすれば、
    #    終局時の情報が得られる
    assert env.done("game")
    assert env.done("hand")
    assert len(obs_dict) == 4
    for _, obs in obs_dict.items():
        assert len(obs.legal_actions()) == 1
        assert obs.legal_actions()[0].type() == mjx.ActionType.DUMMY
    rewards = env.rewards()

    assert len(rewards) == 4
    assert rewards["player_0"] == 90
    assert rewards["player_1"] == 0
    assert rewards["player_2"] == 45
    assert rewards["player_3"] == -135

    # if dummy actions are taken after done("game")
    # 基本的にはdone("game")がTrueになったらresetを呼ぶ想定
    # done("game")がtrueになってdummyをstepで呼んだ後の値は使われる想定ではないが
    # 今の定義は、他のdone("hand")がtrueになってdummyをstepに渡した直後との整合性を取るため、
    # rewardは全て0で、doneもすべてFalse
    # ただ、done()やrewards()が呼ばれたらexceptionを投げるように変更するかもしれない
    action_dict = {}
    for agent, obs in obs_dict.items():
        action_dict[agent] = random_agent.act(obs)
    assert len(action_dict) == 4
    obs_dict = env.step(action_dict)
    assert obs_dict == {}
    # TODO: raise exceptions or print warning to ask user to call reset
    rewards = env.rewards()
    assert len(rewards) == 4
    assert rewards["player_0"] == 0
    assert rewards["player_1"] == 0
    assert rewards["player_2"] == 0
    assert rewards["player_3"] == 0
    assert not env.done()  # done == Trueとなるタイミングは各局一度だけ
    assert not env.done("hand")  # done == Trueとなるタイミングは各局一度だけ

    # test specifying dealer order
    obs_dict = env.reset(1234, ["player_3", "player_1", "player_2", "player_0"])
    assert len(obs_dict) == 1
    assert "player_3" in obs_dict

    obs_dict = env.reset(1234)
    assert len(obs_dict) == 1
    assert "player_2" in obs_dict


def testMjxEnvRewardsHandWin():
    random_agent = mjx.agent.RuleBasedAgent()
    random.seed(1234)
    env = mjx.env.MjxEnv()
    obs_dict = env.reset(1234)
    while not env.done("hand"):
        action_dict = {}
        for agent, obs in obs_dict.items():
            action_dict[agent] = random_agent.act(obs)
        obs_dict = env.step(action_dict)
        if not env.done("hand"):
            rewards = env.rewards("hand_win")
            assert len(rewards) == 4
            assert rewards["player_0"] == 0
            assert rewards["player_1"] == 0
            assert rewards["player_2"] == 0
            assert rewards["player_3"] == 0

    rewards = env.rewards("hand_win")
    assert len(rewards) == 4
    assert rewards["player_0"] == 0, env.state().to_json()
    assert rewards["player_1"] == 0, env.state().to_json()
    assert rewards["player_2"] == 0, env.state().to_json()
    assert rewards["player_3"] == 1, env.state().to_json()


def testMjxEnvRoundDone():
    random_agent = mjx.agent.RandomDebugAgent()
    random.seed(1234)
    env = mjx.env.MjxEnv()
    obs_dict = env.reset(1234)
    while not env.done(done_type="hand"):
        action_dict = {}
        for agent, obs in obs_dict.items():
            action_dict[agent] = random_agent.act(obs)
        obs_dict = env.step(action_dict)
    assert not env.done("game")
    assert env.done("hand")
    assert len(obs_dict) == 4
    for _, obs in obs_dict.items():
        assert len(obs.legal_actions()) == 1
        assert obs.legal_actions()[0].type() == mjx.ActionType.DUMMY



from __future__ import annotations

import json
import sys
from dataclasses import dataclass
from typing import List, Optional, Union

from google.protobuf import json_format
from rich.console import Console
from rich.layout import Layout
from rich.panel import Panel
from rich.text import Text

import mjxproto
from mjx.action import Action
from mjx.const import EventType, RelativePlayerIdx
from mjx.open import Open
from mjx.tile import Tile as _Tile
from mjx.visualizer.converter import get_event_type, get_modifier, get_tile_char, get_wind_char


@dataclass
class GameVisualConfig:
    path: str = "observations.json"
    mode: str = "obs"
    uni: bool = False
    rich: bool = False
    lang: int = 0
    show_name: bool = True


class Tile(_Tile):
    def __init__(
        self,
        tile_id: int,
        is_open: bool = False,
        is_tsumogiri: bool = False,
        with_riichi: bool = False,
        highlighting: bool = False,
    ):
        super().__init__(tile_id)
        self.visual_char: str = "error"  # 実際に表示するときの文字列（裏向き含む）
        self.is_open: bool = is_open  # 表向きか（自分の手牌以外は基本的に裏）
        self.is_tsumogiri: bool = is_tsumogiri  # ツモ切り表示にするかどうか
        self.with_riichi: bool = with_riichi  # 横向きにするかどうか
        self.is_transparent: bool = False  # 鳴かれた牌は透明にして河に表示
        self.is_highlighting: bool = highlighting  # 最後のアクションをハイライト表示


class TileUnit:
    def __init__(
        self,
        tiles_type: EventType,
        from_who: Optional[RelativePlayerIdx],
        tiles: List[Tile],
    ):
        self.tile_unit_type: EventType = tiles_type
        self.from_who: Optional[RelativePlayerIdx] = from_who
        self.tiles: List[Tile] = tiles


class Player:
    def __init__(self, idx: int):
        self.player_idx: int = idx
        self.wind: int = 0
        self.score: str = ""
        self.tile_units: List[TileUnit] = [
            TileUnit(
                EventType.DISCARD,
                None,
                [],
            )
        ]
        self.riichi_now: bool = False
        self.is_declared_riichi: bool = False
        self.name: str = ""
        self.draw_now: bool = False
        self.win_tile_id: int = -1


class MahjongTable:
    """
    MahjongTableクラスは場の情報（プレイヤーの手牌や河など）を保持します。
    """

    def __init__(self):
        self.players = [Player(i) for i in range(4)]
        self.riichi: int = 0
        self.round: int = 0
        self.honba: int = 0
        self.my_idx: int = 0  # The player you want to show.
        self.wall_num: int = 136
        self.doras: List[int] = []
        self.uradoras: List[int] = []
        self.result: str = ""
        self.event_info: Optional[EventType] = None
        self.legal_actions: List[Action] = []
        self.new_dora: int = -1

    def get_wall_num(self) -> int:
        all = 136 - 14
        for p in self.players:
            for t_u in p.tile_units:
                all -= len([tile for tile in t_u.tiles if not tile.is_transparent])
        return all

    def check_num_tiles(self) -> bool:
        for p in self.players:
            num_of_tiles = 0
            hand = ""
            open = []
            for tile_unit in p.tile_units:
                if tile_unit.tile_unit_type == EventType.DRAW:
                    num_of_tiles += len(tile_unit.tiles)
                    hand = " ".join([str(tile.id) for tile in tile_unit.tiles])
                elif tile_unit.tile_unit_type != EventType.DISCARD:
                    num_of_tiles += 3
                    open.append(" ".join([str(tile.id) for tile in tile_unit.tiles]))

            open_str = " : ".join(open)
            if num_of_tiles < 13 or 14 < num_of_tiles:
                sys.stderr.write(
                    f"ERROR: The number of tiles is inaccurate. Player: {p.player_idx}\nhand:[{hand}],open:[{open_str}]"
                )
                return False
        return True

    @staticmethod
    def load_proto_data(path) -> List[Union[mjxproto.Observation, mjxproto.State]]:
        print(path, flush=True)
        with open(path, "r") as f:
            proto_data_list = []
            for line in f:
                line = line.strip("\n").strip()
                proto_data = MahjongTable.json_to_proto(line)
                proto_data_list.append(proto_data)
        return proto_data_list

    @classmethod
    def from_proto(cls, proto_data: Union[mjxproto.Observation, mjxproto.State]):
        if isinstance(proto_data, mjxproto.Observation):
            return cls.decode_observation(proto_data)
        else:
            return cls.decode_state(proto_data)

    @staticmethod
    def json_to_proto(json_data: str) -> Union[mjxproto.State, mjxproto.Observation]:
        try:
            observation = json_format.ParseDict(json.loads(json_data), mjxproto.Observation())
            return observation
        except json_format.ParseError:
            try:
                state = json_format.ParseDict(json.loads(json_data), mjxproto.State())
                return state
            except json_format.ParseError:
                raise ValueError(
                    f"Input json cannot be converted to either State or Observation.\n{json_data}"
                )

    @classmethod
    def decode_private_observation(
        cls, table: MahjongTable, private_observation: mjxproto.PrivateObservation, who: int
    ):
        """
        MahjongTableのデータに、
        手牌の情報を読み込ませる関数
        """
        p = table.players[who]

        has_tsumotile = False
        if private_observation.draw_history != [] and (
            private_observation.draw_history[-1]
            in [id for id in private_observation.curr_hand.closed_tiles]
        ):
            has_tsumotile = True

        if p.draw_now and has_tsumotile:
            p.tile_units.append(
                TileUnit(
                    EventType.DRAW,
                    None,
                    [
                        Tile(i, is_open=True)
                        for i in private_observation.curr_hand.closed_tiles
                        if i != private_observation.draw_history[-1]
                    ]
                    + [
                        Tile(private_observation.draw_history[-1], is_open=True, highlighting=True)
                    ],
                )
            )
        else:
            p.tile_units.append(
                TileUnit(
                    EventType.DRAW,
                    None,
                    [Tile(i, is_open=True) for i in private_observation.curr_hand.closed_tiles],
                )
            )

        return table

    @classmethod
    def decode_public_observation(
        cls, table: MahjongTable, public_observation: mjxproto.PublicObservation
    ):
        """
        MahjongTableのデータに、
        手牌**以外**の情報を読み込ませる関数
        """
        table.round = public_observation.init_score.round + 1
        table.honba = public_observation.init_score.honba
        table.riichi = public_observation.init_score.riichi
        table.doras = list(public_observation.dora_indicators)

        for i, p in enumerate(table.players):
            p.name = public_observation.player_ids[i]
            p.score = str(public_observation.init_score.tens[i])

        for i, eve in enumerate(public_observation.events):
            p = table.players[eve.who]

            p.draw_now = eve.type == EventType.DRAW

            if eve.type == EventType.DISCARD:
                discard = [
                    _discard
                    for _discard in p.tile_units
                    if _discard.tile_unit_type == EventType.DISCARD
                ][0]

                if p.riichi_now:
                    discard.tiles.append(
                        Tile(
                            eve.tile,
                            is_open=True,
                            with_riichi=True,
                        )
                    )
                    p.riichi_now = False
                else:
                    discard.tiles.append(
                        Tile(
                            eve.tile,
                            is_open=True,
                        )
                    )
                if i == len(public_observation.events) - 1:
                    discard.tiles[-1].is_highlighting = True

            elif eve.type == EventType.TSUMOGIRI:
                discard = [
                    _discard
                    for _discard in p.tile_units
                    if _discard.tile_unit_type == EventType.DISCARD
                ][0]

                if p.riichi_now:
                    discard.tiles.append(
                        Tile(
                            eve.tile,
                            is_open=True,
                            is_tsumogiri=True,
                            with_riichi=True,
                        )
                    )
                    p.riichi_now = False
                else:
                    discard.tiles.append(Tile(eve.tile, is_open=True, is_tsumogiri=True))

                if i == len(public_observation.events) - 1:
                    discard.tiles[-1].is_highlighting = True

            elif eve.type == EventType.RIICHI:
                p.riichi_now = True

            elif eve.type == EventType.RIICHI_SCORE_CHANGE:
                table.riichi += 1
                p.is_declared_riichi = True
                p.score = str(int(p.score) - 1000)
                if i == len(public_observation.events) - 1:
                    discard = [
                        _discard
                        for _discard in p.tile_units
                        if _discard.tile_unit_type == EventType.DISCARD
                    ][0]
                    discard.tiles[-1].is_highlighting = True

            elif eve.type == EventType.RON:
                if public_observation.events[i - 1].type != EventType.RON:
                    _p = table.players[public_observation.events[i - 1].who]
                    discard = [
                        _discard
                        for _discard in _p.tile_units
                        if _discard.tile_unit_type == EventType.DISCARD
                    ][0]
                    discard.tiles[-1].is_transparent = True

            elif eve.type in [
                EventType.CHI,
                EventType.PON,
                EventType.CLOSED_KAN,
                EventType.ADDED_KAN,
                EventType.OPEN_KAN,
            ]:
                open_data = Open(eve.open)
                open_tiles = open_data.tiles()

                if eve.type == EventType.ADDED_KAN:
                    # added_kanのときにすでに存在するポンを取り除く処理
                    p.tile_units = [
                        t_u
                        for t_u in p.tile_units
                        if (
                            t_u.tile_unit_type != EventType.PON
                            or t_u.tiles[0].id() // 4 != open_tiles[0].id() // 4
                        )
                    ]

                # 鳴き牌を追加する処理
                if eve.type in [
                    EventType.CLOSED_KAN,
                    EventType.ADDED_KAN,
                ]:
                    p.tile_units.append(
                        TileUnit(
                            open_data.event_type(),
                            open_data.steal_from(),
                            [Tile(i.id(), is_open=True) for i in open_tiles],
                        )
                    )
                    if i == len(public_observation.events) - 1:
                        p.tile_units[-1].tiles[0].is_highlighting = True

                # 鳴かれた牌を透明にする処理
                if eve.type in [
                    EventType.CHI,
                    EventType.PON,
                    EventType.OPEN_KAN,
                ]:
                    idx_from = p.player_idx
                    if open_data.steal_from() == RelativePlayerIdx.LEFT:
                        idx_from = (p.player_idx + 3) % 4
                    elif open_data.steal_from() == RelativePlayerIdx.CENTER:
                        idx_from = (p.player_idx + 2) % 4
                    elif open_data.steal_from() == RelativePlayerIdx.RIGHT:
                        idx_from = (p.player_idx + 1) % 4

                    p_from = table.players[idx_from]
                    for p_from_t_u in p_from.tile_units:
                        if p_from_t_u.tile_unit_type == EventType.DISCARD:
                            p_from_t_u.tiles[-1].is_transparent = True

                            # 鳴き牌を追加する処理
                            p.tile_units.append(
                                TileUnit(
                                    open_data.event_type(),
                                    open_data.steal_from(),
                                    [
                                        Tile(
                                            p_from_t_u.tiles[-1].id(),
                                            is_open=True,
                                            highlighting=(
                                                i == len(public_observation.events) - 1
                                            ),  # 最新のアクションならハイライト
                                        )
                                    ]
                                    + [
                                        Tile(i.id(), is_open=True)
                                        for i in open_tiles
                                        if i != p_from_t_u.tiles[-1].id()
                                    ],
                                )
                            )
                            assert p.tile_units[-1].tiles[0].id() == p_from_t_u.tiles[-1].id()

            elif eve.type == EventType.NEW_DORA:
                table.new_dora = eve.tile

        if len(public_observation.events) == 0:
            return table

        last_event = public_observation.events[-1]

        if last_event.type in [
            EventType.TSUMO,
            EventType.RON,
        ]:
            table.result = "win"
            table.event_info = last_event.type
            table.players[last_event.who].win_tile_id = last_event.tile
        elif last_event.type in [
            EventType.ABORTIVE_DRAW_NINE_TERMINALS,
            EventType.ABORTIVE_DRAW_FOUR_RIICHIS,
            EventType.ABORTIVE_DRAW_THREE_RONS,
            EventType.ABORTIVE_DRAW_FOUR_KANS,
            EventType.ABORTIVE_DRAW_FOUR_WINDS,
            EventType.ABORTIVE_DRAW_NORMAL,
            EventType.ABORTIVE_DRAW_NAGASHI_MANGAN,
        ]:
            table.result = "nowinner"
            table.event_info = last_event.type
            discard = [
                _discard
                for _discard in table.players[public_observation.events[-2].who].tile_units
                if _discard.tile_unit_type == EventType.DISCARD
            ][0]
            discard.tiles[-1].is_highlighting = True

        return table

    @classmethod
    def decode_round_terminal(cls, table: MahjongTable, round_terminal, win: bool):
        final_ten_changes = [0, 0, 0, 0]
        if win:
            for win_data in round_terminal.wins:
                table.uradoras = win_data.ura_dora_indicators
                for i in range(4):
                    p = table.players[i]
                    if p.player_idx == win_data.who:
                        # 手牌をround_terminalのもので上書き
                        p.tile_units = [
                            t_u for t_u in p.tile_units if t_u.tile_unit_type != EventType.DRAW
                        ]

                        p.tile_units.append(
                            TileUnit(
                                EventType.DRAW,
                                None,
                                [
                                    Tile(
                                        i,
                                        is_open=True,
                                        highlighting=(i == p.win_tile_id),
                                    )
                                    for i in win_data.hand.closed_tiles
                                ],
                            )
                        )

                    final_ten_changes[i] += win_data.ten_changes[i]

            # 複数人勝者が居た場合のために、後でまとめて点数移動を計算
            for i, p in enumerate(table.players):
                delta = final_ten_changes[i]
                p.score = (
                    str(int(p.score) + delta) + ("(+" if delta > 0 else "(") + str(delta) + ")"
                )

        else:  # nowinner
            for tenpai_data in round_terminal.no_winner.tenpais:
                for i in range(4):
                    if table.players[i].player_idx == tenpai_data.who:
                        # 手牌をround_terminalのもので上書き
                        table.players[i].tile_units = [
                            t_u
                            for t_u in table.players[i].tile_units
                            if t_u.tile_unit_type != EventType.DRAW
                        ]

                        table.players[i].tile_units.append(
                            TileUnit(
                                EventType.DRAW,
                                None,
                                [Tile(i, is_open=True) for i in tenpai_data.hand.closed_tiles],
                            )
                        )

            for i, p in enumerate(table.players):
                delta = round_terminal.no_winner.ten_changes[i]
                p.score = (
                    str(int(p.score) + delta) + ("(+" if delta > 0 else "(") + str(delta) + ")"
                )

        return table

    @classmethod
    def decode_observation(cls, proto_data: mjxproto.Observation):
        table = MahjongTable()

        table.my_idx = proto_data.who
        table = cls.decode_public_observation(table, proto_data.public_observation)
        table = cls.decode_private_observation(
            table, proto_data.private_observation, proto_data.who
        )

        # Obsevationの場合,適切な数の裏向きの手牌を用意する
        for i, p in enumerate(table.players):
            if p.player_idx != table.my_idx:
                hands_num = 13
                for t_u in p.tile_units:
                    if t_u.tile_unit_type not in [
                        EventType.DISCARD,
                        EventType.DRAW,
                    ]:
                        hands_num -= 3

                p.tile_units.append(
                    TileUnit(
                        EventType.DRAW,
                        None,
                        [Tile(i, is_open=False) for i in range(hands_num)],
                    )
                )
            p.wind = (-table.round + 1 + i) % 4

        table.wall_num = cls.get_wall_num(table)
        table.legal_actions = [Action.from_proto(act) for act in proto_data.legal_actions]
        table.legal_actions.sort(key=lambda a: a.to_idx())

        if len(proto_data.public_observation.events) != 0:
            if table.result == "win":
                table = cls.decode_round_terminal(table, proto_data.round_terminal, True)

            elif table.result == "nowinner":
                table = cls.decode_round_terminal(table, proto_data.round_terminal, False)

        return table

    @classmethod
    def decode_state(cls, proto_data: mjxproto.State):
        table = MahjongTable()

        table = cls.decode_public_observation(table, proto_data.public_observation)
        for i in range(4):
            table = cls.decode_private_observation(table, proto_data.private_observations[i], i)

        for i, p in enumerate(table.players):
            p.wind = (-table.round + 1 + i) % 4

        table.wall_num = cls.get_wall_num(table)

        if len(proto_data.public_observation.events) != 0:
            if table.result == "win":
                table = cls.decode_round_terminal(table, proto_data.round_terminal, True)

            elif table.result == "nowinner":
                table = cls.decode_round_terminal(table, proto_data.round_terminal, False)

        return table


class GameBoardVisualizer:
    """
    GameBoardVisualizer クラスは内部にMahjongTableクラスのオブジェクトを持ち、
    その表示などを行います。
    """

    def __init__(self, config: GameVisualConfig):
        self.config = config

    def get_layout(self):
        layout = Layout()
        if self.config.show_name:
            layout.split_column(
                Layout(" ", name="space_top"),
                Layout(name="info"),
                Layout(name="players_info_top"),
                Layout(name="table"),
            )
            layout["players_info_top"].size = 7
            layout["players_info_top"].split_row(
                Layout(" ", name="player1_info_top"),
                Layout(" ", name="player2_info_top"),
                Layout(" ", name="player3_info_top"),
                Layout(" ", name="player4_info_top"),
            )
        else:
            layout.split_column(
                Layout(" ", name="space_top"),
                Layout(name="info"),
                Layout(name="table"),
            )

        layout["space_top"].size = 3
        layout["info"].size = 3
        layout["table"].minimum_size = 20

        layout["table"].split_column(
            Layout(name="upper1"),
            Layout(name="middle1"),
            Layout(name="lower1"),
        )

        layout["upper1"].size = 3
        layout["lower1"].size = 3

        layout["upper1"].split_row(
            Layout(" "),
            Layout(" ", name="hand3"),
            Layout(" "),
        )
        layout["hand3"].ratio = 6

        layout["middle1"].split_row(
            Layout(" ", name="hand4"),
            Layout(name="middle2"),
            Layout(" ", name="hand2"),
        )
        layout["middle2"].ratio = 10

        layout["lower1"].split_row(
            Layout(" "),
            Layout(" ", name="hand1"),
            Layout(" "),
        )
        layout["hand1"].ratio = 6

        layout["middle2"].split_column(
            Layout(name="upper2"),
            Layout(name="middle3"),
            Layout(name="lower2"),
        )

        layout["upper2"].split_row(
            Layout(" ", name="player3_info_corner"),
            Layout(" ", name="discard3"),
            Layout(" ", name="player2_info_corner"),
        )

        layout["middle3"].split_row(
            Layout(" ", name="discard4"),
            Layout(" ", name="middle4"),
            Layout(" ", name="discard2"),
        )

        layout["middle4"].split_column(
            Layout(" ", name="space_for_info_center"),
            Layout(" ", name="player3_info_center"),
            Layout(" ", name="middle5"),
            Layout(" ", name="player1_info_center"),
        )
        layout["space_for_info_center"].size = 1

        layout["middle5"].split_row(
            Layout(" ", name="player4_info_center"),
            Layout(" ", name="player2_info_center"),
        )
        layout["lower2"].split_row(
            Layout(" ", name="player4_info_corner"),
            Layout(" ", name="discard1"),
            Layout(" ", name="player1_info_corner"),
        )

        return layout

    def add_suffix(self, tile_unit: TileUnit, player_idx: int = 0) -> str:
        """
        wdwdwd -> wdwdwdL
        のように、情報を付け加える関数。
        """

        for tile in tile_unit.tiles:
            if not tile.is_open:
                tile.visual_char = "\U0001F02B" if self.config.uni else "#"
            else:
                tile.visual_char = get_tile_char(tile.id(), self.config.uni)
            if tile.is_tsumogiri:
                if tile.with_riichi:
                    if self.config.uni and tile.visual_char != "\U0001F004\uFE0E":
                        tile.visual_char += " *r"
                    else:
                        tile.visual_char += "*r"
                else:
                    if self.config.uni and tile.visual_char != "\U0001F004\uFE0E":
                        tile.visual_char += " *"
                    else:
                        tile.visual_char += "*"
            elif tile.with_riichi:
                if self.config.uni and tile.visual_char != "\U0001F004\uFE0E":
                    tile.visual_char += " r"
                else:
                    tile.visual_char += "r"

        if self.config.rich:
            if tile_unit.tile_unit_type == EventType.DISCARD:
                discards = [
                    tile.visual_char
                    + (
                        ""
                        if (
                            tile.visual_char == "\U0001F004\uFE0E"
                            or (tile.is_tsumogiri and tile.with_riichi)
                        )
                        else " "
                    )
                    + (
                        ""
                        if (tile.is_tsumogiri or tile.with_riichi)
                        else "  "
                        if self.config.uni
                        else " "
                    )
                    for tile in tile_unit.tiles
                ]
                tiles = "\n".join(
                    ["".join(discards[idx : idx + 6]) for idx in range(0, len(discards), 6)]
                )
                return tiles
            elif player_idx == 1:
                tiles = (
                    "\n"
                    + get_modifier(tile_unit.from_who, tile_unit.tile_unit_type)
                    + "\n"
                    + "\n".join(
                        [
                            (" " if tile.visual_char == "\U0001F004\uFE0E" else "")
                            + tile.visual_char
                            for tile in tile_unit.tiles
                        ]
                    )
                    + "\n"
                )
                return tiles
            elif player_idx == 2:
                if tile_unit.tile_unit_type == EventType.DRAW:
                    tiles = "".join(
                        [
                            tile.visual_char
                            + ("" if tile.visual_char == "\U0001F004\uFE0E" else " ")
                            for tile in tile_unit.tiles
                        ]
                    )
                    return tiles
                tiles = get_modifier(tile_unit.from_who, tile_unit.tile_unit_type) + "".join(
                    [
                        tile.visual_char
                        + (
                            ""
                            if (not self.config.uni or tile.visual_char == "\U0001F004\uFE0E")
                            else " "
                        )
                        for tile in sorted(
                            tile_unit.tiles,
                            key=lambda x: x.id(),
                            reverse=True,
                        )
                    ]
                )
                return tiles
            elif player_idx == 3:
                tiles = (
                    "\n"
                    + "\n".join(
                        [
                            (" " if tile.visual_char == "\U0001F004\uFE0E" else "")
                            + tile.visual_char
                            for tile in tile_unit.tiles
                        ]
                    )
                    + "\n "
                    + get_modifier(tile_unit.from_who, tile_unit.tile_unit_type)
                    + "\n"
                )
                return tiles
            else:
                if tile_unit.tile_unit_type == EventType.DRAW:
                    tiles = "".join(
                        [
                            tile.visual_char
                            + ("" if tile.visual_char == "\U0001F004\uFE0E" else " ")
                            for tile in tile_unit.tiles
                        ]
                    )
                    return tiles
                tiles = (
                    "".join(
                        [
                            tile.visual_char
                            + (
                                ""
                                if (not self.config.uni or tile.visual_char == "\U0001F004\uFE0E")
                                else " "
                            )
                            for tile in tile_unit.tiles
                        ]
                    )
                    + get_modifier(tile_unit.from_who, tile_unit.tile_unit_type)
                    + " "
                )
                return tiles
        else:  # not rich
            if tile_unit.tile_unit_type == EventType.DRAW:
                tiles = "".join(
                    [
                        tile.visual_char + ("" if tile.visual_char == "\U0001F004\uFE0E" else " ")
                        for tile in tile_unit.tiles
                    ]
                )
                return tiles
            if tile_unit.tile_unit_type == EventType.DISCARD:
                discards = [
                    tile.visual_char
                    + (
                        ""
                        if (
                            tile.visual_char == "\U0001F004\uFE0E"
                            or (tile.is_tsumogiri and tile.with_riichi)
                        )
                        else " "
                    )
                    + (
                        ""
                        if (tile.is_tsumogiri or tile.with_riichi)
                        else "  "
                        if self.config.uni
                        else " "
                    )
                    for tile in tile_unit.tiles
                ]
                tiles = "\n".join(
                    ["".join(discards[idx : idx + 6]) for idx in range(0, len(discards), 6)]
                )
                return tiles

            if tile_unit.tile_unit_type in [
                EventType.CHI,
                EventType.PON,
                EventType.CLOSED_KAN,
                EventType.OPEN_KAN,
                EventType.ADDED_KAN,
            ]:
                tiles = "".join(
                    [
                        tile.visual_char
                        + (
                            ""
                            if (not self.config.uni or tile.visual_char == "\U0001F004\uFE0E")
                            else " "
                        )
                        for tile in tile_unit.tiles
                    ]
                ) + get_modifier(tile_unit.from_who, tile_unit.tile_unit_type)
                return tiles
        return "error"

    def get_board_info(self, table: MahjongTable) -> str:
        board_info = []
        board_info.append(
            [
                f"round:{table.round}",
                get_wind_char((table.round - 1) // 4, self.config.lang)
                + str((table.round - 1) % 4 + 1)
                + "局",
            ][self.config.lang]
        )
        if table.honba > 0:
            board_info.append(
                " " + ["honba:" + str(table.honba), str(table.honba) + "本場"][self.config.lang]
            )
        if table.riichi > 0:
            board_info.append(" " + ["riichi:", "供託"][self.config.lang] + str(table.riichi))
        board_info.append(
            " "
            + ["wall:" + str(table.wall_num), "残り:" + str(table.wall_num) + "枚"][self.config.lang]
        )
        dora = "".join(
            [
                get_tile_char(d, self.config.uni)
                + ("" if get_tile_char(d, self.config.uni) == "\U0001F004\uFE0E" else " ")
                for d in table.doras
            ]
        )
        board_info.append(" " + ["Dora:", "ドラ:"][self.config.lang] + dora)
        uradora = "".join(
            [
                get_tile_char(d, self.config.uni)
                + ("" if get_tile_char(d, self.config.uni) == "\U0001F004\uFE0E" else " ")
                for d in table.uradoras
            ]
        )
        if uradora != "":
            board_info.append(" " + ["UraDora:", "裏ドラ:"][self.config.lang] + uradora)

        if table.event_info is not None:
            event_info = get_event_type(table.event_info, self.config.lang)
            board_info.append("    " + event_info)

        return "".join(board_info)

    def get_text_width(self, text: str):
        import unicodedata

        text_counter: int = 0
        for c in text:
            if unicodedata.east_asian_width(c) in "FWA":
                text_counter = text_counter + 2
            else:
                text_counter = text_counter + 1
        return text_counter

    def show_by_text(self, table: MahjongTable) -> str:
        board_info = self.get_board_info(table)
        board_info = (
            "#" * (self.get_text_width(board_info) + 2)
            + "\n"
            + "#"
            + board_info
            + "#\n"
            + "#" * (self.get_text_width(board_info) + 2)
            + "\n"
        )

        players_info = []
        table.players.sort(key=lambda x: (x.player_idx - table.my_idx) % 4)
        for i, p in enumerate(table.players):
            player_info = []

            player_info.append(
                get_wind_char(p.wind, self.config.lang)
                + " [ "
                + "".join(
                    [
                        p.score
                        + ([", riichi", ", リーチ"][self.config.lang] if p.is_declared_riichi else "")
                    ]
                )
                + " ]",
            )
            if self.config.show_name:
                player_info.append(" " + p.name)
            player_info.append("\n\n")

            opens = []
            hand = ""
            discards = ""
            for t_u in reversed(p.tile_units):
                if t_u.tile_unit_type == EventType.DRAW:
                    hand = self.add_suffix(t_u)
                elif t_u.tile_unit_type == EventType.DISCARD:
                    discards = self.add_suffix(t_u)
                else:
                    opens.append(self.add_suffix(t_u))

            hand_area = hand + "      " + " ".join(opens)
            player_info.append(hand_area)
            player_info.append("\n\n")

            player_info.append(discards)
            player_info.append("\n\n\n")
            players_info.append("".join(player_info))

        return board_info + "".join(players_info)

    def show_by_rich(self, table: MahjongTable) -> None:
        layout = self.get_layout()

        layout["info"].update(
            Panel(
                Text(self.get_board_info(table), justify="center", style="color(1)"),
                style="bold green",
            )
        )

        table.players.sort(key=lambda x: (x.player_idx - table.my_idx) % 4)

        players_info_top = [
            "player1_info_top",
            "player2_info_top",
            "player3_info_top",
            "player4_info_top",
        ]
        players_info_center = [
            "player1_info_center",
            "player2_info_center",
            "player3_info_center",
            "player4_info_center",
        ]
        hands_idx = ["hand1", "hand2", "hand3", "hand4"]
        discards_idx = [
            "player1_discard",
            "player2_discard",
            "player3_discard",
            "player4_discard",
        ]
        discards_idx = [
            "discard1",
            "discard2",
            "discard3",
            "discard4",
        ]

        for i, p in enumerate(table.players):
            wind = Text(
                get_wind_char(p.wind, self.config.lang),
                justify="center",
                style="bold green",
            )

            score = Text(p.score, justify="center", style="yellow")

            riichi = Text()
            if p.is_declared_riichi:
                riichi = [
                    Text(" riichi", style="yellow"),
                    Text(" リーチ", style="yellow"),
                ][self.config.lang]

            player_info = wind + Text("\n") + score + riichi

            layout[players_info_center[i]].update(player_info)

            name = Text(justify="center", style="bold green")
            if self.config.show_name:
                name += Text(" " + p.name, style="white")
                layout[players_info_top[i]].update(
                    Panel(player_info + Text("\n\n") + name, style="bold green")
                )

            opens = []
            hand = ""
            discards = ""
            for t_u in reversed(p.tile_units):
                if t_u.tile_unit_type == EventType.DRAW:
                    hand = self.add_suffix(t_u, player_idx=i)
                elif t_u.tile_unit_type == EventType.DISCARD:
                    discards = self.add_suffix(t_u, player_idx=i)
                else:
                    opens.append(self.add_suffix(t_u, player_idx=i))
            if p.player_idx in [(table.my_idx + 1) % 4, (table.my_idx + 2) % 4]:
                hand_area = " ".join(opens) + "      " + hand
            else:
                hand_area = hand + "      " + " ".join(opens)
            layout[hands_idx[i]].update(
                Panel(
                    Text(hand_area, justify="center", no_wrap=True, style="white"),
                    style="bold green",
                )
            )

            layout[discards_idx[i]].update(
                Panel(
                    Text(
                        discards,
                        justify="left",
                        style="white",
                    ),
                    style="bold green",
                )
            )

        console = Console()
        console.print(layout)

    def print(self, data: MahjongTable):
        if self.config.rich:
            self.show_by_rich(data)
        else:
            print(self.show_by_text(data))



import base64
from importlib.resources import read_binary
from typing import List, Optional, Tuple, Union

import svgwrite
from svgwrite.drawing import Drawing

import mjx.visualizer
import mjxproto
from mjx.const import EventType as TileUnitType
from mjx.const import RelativePlayerIdx
from mjx.visualizer.visualizer import MahjongTable, Tile, get_tile_char, get_wind_char


def dwg_add(
    dwg_p,
    dwg_g,
    pos,
    txt: str,
    rotate: bool = False,
    is_red: bool = False,
    transparent: bool = False,
    highliting: bool = False,
):
    opacity = 1.0
    if transparent:
        opacity = 0.5

    highlight_fill = "black"
    highlight_opacity = 1.0
    highlight_stroke_width = 2.0

    if rotate:
        if is_red:
            horizontal_tiles = [
                dwg_p.text(txt[0], insert=(0, 0), fill="red", opacity=opacity),
                dwg_p.text(
                    "\U0001F006",
                    insert=(0, 0),
                    stroke=svgwrite.rgb(255, 255, 255, "%"),
                    fill="white",
                ),
                dwg_p.text("\U0001F006", insert=(0, 0), fill="black", opacity=opacity),
            ]

            for horizontal_tile in horizontal_tiles:
                horizontal_tile.rotate(90, (0, 0))
                horizontal_tile.translate(pos)
                dwg_g.add(horizontal_tile)

        else:
            horizontal_tile = dwg_p.text(txt[0], insert=(0, 0), opacity=opacity)
            horizontal_tile.rotate(90, (0, 0))
            horizontal_tile.translate(pos)
            dwg_g.add(horizontal_tile)

        if highliting:
            highlighted_tile = dwg_p.text(
                "\U0001F006",
                insert=(0, 0),
                stroke=svgwrite.rgb(0, 0, 0, "%"),
                stroke_width=highlight_stroke_width,
                fill=highlight_fill,
                opacity=highlight_opacity,
            )
            highlighted_tile.rotate(90, (0, 0))
            highlighted_tile.translate(pos)
            dwg_g.add(highlighted_tile)

    else:
        if is_red:
            dwg_g.add(dwg_p.text(txt[0], pos, fill="red", opacity=opacity))
            dwg_g.add(
                dwg_p.text(
                    "\U0001F006",
                    pos,
                    stroke=svgwrite.rgb(255, 255, 255, "%"),
                    fill="white",
                    opacity=opacity,
                )
            )
            dwg_g.add(dwg_p.text("\U0001F006", pos, fill="black", opacity=opacity))
        else:
            dwg_g.add(dwg_p.text(txt[0], pos, opacity=opacity))

        if highliting:
            dwg_g.add(
                dwg_p.text(
                    "\U0001F006",
                    pos,
                    stroke=svgwrite.rgb(0, 0, 0, "%"),  # 255,140,0
                    stroke_width=highlight_stroke_width,
                    fill=highlight_fill,
                    opacity=highlight_opacity,
                )
            )


def _make_svg(
    proto_data: Union[mjxproto.State, mjxproto.Observation],
    filename: str,
    target_idx: Optional[int] = None,
    show_name: bool = False,
    highlight_last_event: bool = False,
) -> Drawing:
    assert filename.endswith(".svg")

    sample_data: MahjongTable
    if isinstance(proto_data, mjxproto.Observation):
        sample_data = MahjongTable.decode_observation(proto_data)
    else:
        sample_data = MahjongTable.decode_state(proto_data)

    width = 800
    height = 800
    char_width = 32  # 45:28.8,60:38.4
    char_height = 44  # 45:40.5,60:53
    red_hai = [16, 52, 88]

    if target_idx is None:
        if isinstance(proto_data, mjxproto.Observation):
            target_idx = proto_data.who
        else:
            target_idx = 0
    assert target_idx is not None
    sample_data.players.sort(key=lambda x: (x.player_idx - target_idx) % 4)

    dwg = svgwrite.Drawing(
        filename,
        (width, height),
        debug=True,
    )

    dwg._embed_font_data(
        "GL-MahjongTile",
        read_binary(mjx.visualizer, "GL-MahjongTile.ttf"),
        "application/x-font-ttf",
    )

    player_g = dwg.g()

    players: List[Drawing] = [dwg.g(), dwg.g(), dwg.g(), dwg.g()]
    pai: List[Drawing] = [dwg.g(), dwg.g(), dwg.g(), dwg.g()]
    player_info: List[Drawing] = [dwg.g(), dwg.g(), dwg.g(), dwg.g()]
    winds: List[str] = ["", "", "", ""]
    scores: List[str] = ["", "", "", ""]
    is_riichi = [False, False, False, False]

    # Tuple[char, is_red, Tile]
    hands: List[List[Tuple[str, bool, Tile]]] = [[], [], [], []]
    open_tiles: List[
        List[Tuple[List[Tuple[str, bool, Tile]], Optional[RelativePlayerIdx], TileUnitType]]
    ] = [
        [],
        [],
        [],
        [],
    ]
    discards: List[List[Tuple[str, bool, Tile]]] = [[], [], [], []]

    for i in range(4):  # iは各プレイヤー(0-3)
        players[i] = dwg.g()
        pai[i] = dwg.g(style="font-size:50px;font-family:GL-MahjongTile;")
        player_info[i] = dwg.g()

        winds[i] = get_wind_char(sample_data.players[i].wind, lang=1)
        scores[i] = sample_data.players[i].score
        is_riichi[i] = sample_data.players[i].is_declared_riichi

        for t_u in reversed(sample_data.players[i].tile_units):
            if t_u.tile_unit_type == TileUnitType.DRAW:
                for tile in t_u.tiles:
                    hands[i].append(
                        (
                            "\U0001F02B" if not tile.is_open else get_tile_char(tile.id(), True),
                            tile.id in red_hai,
                            tile,
                        )
                    )

            if t_u.tile_unit_type == TileUnitType.DISCARD:
                for tile in t_u.tiles:
                    discards[i].append((get_tile_char(tile.id(), True), tile.id in red_hai, tile))
            if t_u.tile_unit_type in [
                TileUnitType.CHI,
                TileUnitType.PON,
                TileUnitType.CLOSED_KAN,
                TileUnitType.OPEN_KAN,
                TileUnitType.ADDED_KAN,
            ]:
                open_tiles[i].append(
                    (
                        [
                            (get_tile_char(tile.id(), True), tile.id in red_hai, tile)
                            for tile in t_u.tiles
                        ],
                        t_u.from_who,
                        t_u.tile_unit_type,
                    )
                )
    dwg.add(dwg.rect(insert=(0, 0), size=(width, height)))
    dwg.add(dwg.rect(insert=(1, 1), size=(width - 2, height - 2), fill="rgb(255,255,255)"))

    dwg.add(dwg.rect(insert=(278, 278), size=(244, 244)))
    dwg.add(dwg.rect(insert=(279, 279), size=(242, 242), fill="rgb(255,255,255)"))

    # board_info
    if sample_data.honba > 0:
        round = (
            get_wind_char((sample_data.round - 1) // 4, lang=1)
            + str((sample_data.round - 1) % 4 + 1)
            + "局"
        )
        dwg.add(dwg.text(round, (340, 360), style="font-size:24px;font-family:serif;"))
        honba = str(sample_data.honba) + "本場"
        dwg.add(dwg.text(honba, (400, 360), style="font-size:24px;font-family:serif;"))
    else:
        round = (
            get_wind_char((sample_data.round - 1) // 4, lang=1)
            + str((sample_data.round - 1) % 4 + 1)
            + "局"
        )
        dwg.add(dwg.text(round, (368, 360), style="font-size:26px;font-family:serif;"))

    # dora
    doras = [get_tile_char(tile, True) for tile in sample_data.doras]
    while len(doras) < 5:
        doras.append("\U0001F02B")
    dwg.add(
        dwg.text("".join(doras), (337, 400), style="font-size:40px;font-family:GL-MahjongTile;")
    )
    for i, dora in enumerate(sample_data.doras):
        if dora == sample_data.new_dora:
            dwg.add(
                dwg.text(
                    "\U0001F006",
                    (337 + i * 25.6, 400),
                    style="font-size:40px;font-family:GL-MahjongTile;",
                    stroke=svgwrite.rgb(0, 0, 0, "%"),
                    stroke_width=2.0,
                )
            )

    # bou
    b64_1000_mini = base64.b64encode(
        read_binary(mjx.visualizer, "1000_mini.svg"),
    )
    thousand_mini_img = dwg.image("data:image/svg+xml;base64," + b64_1000_mini.decode("ascii"))
    thousand_mini_img.translate(335, 405)
    thousand_mini_img.scale(0.15)
    dwg.add(thousand_mini_img)
    dwg.add(
        dwg.text(
            f"×{sample_data.riichi}",
            (355, 430),
            style="font-size:22px;font-family:serif;",
        )
    )

    b64_hundred_mini = base64.b64encode(
        read_binary(mjx.visualizer, "100_mini.svg"),
    )
    hundred_mini_img = dwg.image("data:image/svg+xml;base64," + b64_hundred_mini.decode("ascii"))
    hundred_mini_img.translate(405, 405)
    hundred_mini_img.scale(0.15)
    dwg.add(hundred_mini_img)
    dwg.add(
        dwg.text(f"×{sample_data.honba}", (425, 430), style="font-size:22px;font-family:serif;")
    )

    # yama_nokori
    wall_num = sample_data.wall_num
    dwg.add(
        dwg.text(
            "\U0001F02B",
            (370, 465),
            style="font-size:30px;font-family:GL-MahjongTile;",
        )
    )
    dwg.add(
        dwg.text(
            "×" + str(wall_num),
            (390, 463),
            style="font-size:22px;font-family:serif;",
        )
    )

    for i in range(4):

        # name
        if show_name:
            player_info[i].add(
                dwg.text(
                    sample_data.players[i].name,
                    (10, 790),
                    style="font-size:18px;font-family:serif;",
                )
            )

        # wind
        player_info[i].add(
            dwg.text(winds[i], (280, 515), style="font-size:30px;font-family:serif;")
        )

        # score
        scores[i] = scores[i].replace("(", " (")
        score_x = width / 2 - len(scores[i]) / 2 * 11
        player_info[i].add(
            dwg.text(scores[i], (score_x, 490), style="font-size:20px;font-family:serif;")
        )

        # riichi_bou
        if is_riichi[i]:
            b64_thousand = base64.b64encode(
                read_binary(mjx.visualizer, "1000.svg"),
            )
            thousand_img = dwg.image("data:image/svg+xml;base64," + b64_thousand.decode("ascii"))
            thousand_img.translate(476, 485)
            thousand_img.scale(0.4)
            thousand_img.rotate(90)
            player_info[i].add(thousand_img)

        left_margin = 190
        # hand
        for j, hand in enumerate(hands[i]):
            hand_txt = hand[0]
            dwg_add(
                dwg,
                pai[i],
                (left_margin + j * char_width, 770),
                hand_txt,
                highliting=hand[2].is_highlighting and highlight_last_event,
            )

        # discard
        riichi_idx = 100000
        for j, discard in enumerate(discards[i]):
            discard_txt = discard[0]
            if discard[2].with_riichi:  # riichi
                riichi_idx = j
                dwg_add(
                    dwg,
                    pai[i],
                    (535 + (j // 6) * char_height, -307 - (j % 6) * char_width),
                    discard_txt,
                    rotate=True,
                    is_red=discard[1],
                    transparent=discard[2].is_transparent,  # 鳴かれた
                    highliting=discard[2].is_highlighting and highlight_last_event,
                )

                if discard[2].is_tsumogiri:
                    dwg_add(
                        dwg,
                        pai[i],
                        (535 + (j // 6) * char_height, -307 - (j % 6) * char_width),
                        "\U0001F02B",
                        rotate=True,
                        transparent=discard[2].is_transparent,
                        highliting=discard[2].is_highlighting and highlight_last_event,
                    )

            elif (riichi_idx < j) and (j // 6 == riichi_idx // 6):
                dwg_add(
                    dwg,
                    pai[i],
                    (
                        304 + char_height - char_width + (j % 6) * char_width,
                        570 + (j // 6) * char_height,
                    ),
                    discard_txt,
                    is_red=discard[1],
                    transparent=discard[2].is_transparent,
                    highliting=discard[2].is_highlighting and highlight_last_event,
                )

                if discard[2].is_tsumogiri:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            304 + char_height - char_width + (j % 6) * char_width,
                            570 + (j // 6) * char_height,
                        ),
                        "\U0001F02B",
                        transparent=discard[2].is_transparent,
                        highliting=discard[2].is_highlighting and highlight_last_event,
                    )
            else:
                dwg_add(
                    dwg,
                    pai[i],
                    (304 + (j % 6) * char_width, 570 + (j // 6) * char_height),
                    discard_txt,
                    is_red=discard[1],
                    transparent=discard[2].is_transparent,
                    highliting=discard[2].is_highlighting and highlight_last_event,
                )

                if discard[2].is_tsumogiri:
                    dwg_add(
                        dwg,
                        pai[i],
                        (304 + (j % 6) * char_width, 570 + (j // 6) * char_height),
                        "\U0001F02B",
                        transparent=discard[2].is_transparent,
                        highliting=discard[2].is_highlighting and highlight_last_event,
                    )

        num_of_tehai = len(hands[i]) + len(open_tiles[i]) * 3
        left_x = char_width if num_of_tehai == 13 else 0

        for open_tile in open_tiles[i]:
            if open_tile[2] == TileUnitType.CHI:
                chi = open_tile[0]
                dwg_add(
                    dwg,
                    pai[i],
                    (
                        741.3,
                        -left_margin  # 初期位置
                        - 3  # 回転時のずれ
                        - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                        - left_x,  # 他の鳴き牌の分のずれ
                    ),
                    chi[0][0],
                    rotate=True,
                    is_red=chi[0][1],
                    highliting=chi[0][2].is_highlighting and highlight_last_event,
                )
                dwg_add(
                    dwg,
                    pai[i],
                    (
                        left_margin + (len(hands[i]) + 1) * char_width + left_x + char_height,
                        770,
                    ),
                    chi[1][0],
                    is_red=chi[1][1],
                    highliting=chi[0][2].is_highlighting and highlight_last_event,
                )
                dwg_add(
                    dwg,
                    pai[i],
                    (
                        left_margin
                        + (len(hands[i]) + 1) * char_width
                        + left_x
                        + char_height
                        + char_width,
                        770,
                    ),
                    chi[2][0],
                    is_red=chi[2][1],
                    highliting=chi[0][2].is_highlighting and highlight_last_event,
                )

                left_x += char_width * 2 + char_height

            elif open_tile[2] == TileUnitType.PON:
                pon = open_tile[0]
                if open_tile[1] == RelativePlayerIdx.LEFT:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            741.3,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x,
                        ),
                        pon[0][0],
                        rotate=True,
                        is_red=pon[0][1],
                        highliting=pon[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x + char_height,
                            770,
                        ),
                        pon[1][0],
                        is_red=pon[1][1],
                        highliting=pon[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin
                            + (len(hands[i]) + 1) * char_width
                            + left_x
                            + char_height
                            + char_width,
                            770,
                        ),
                        pon[2][0],
                        is_red=pon[2][1],
                        highliting=pon[0][2].is_highlighting and highlight_last_event,
                    )

                elif open_tile[1] == RelativePlayerIdx.CENTER:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            741.3,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x
                            - char_width,  # 1つ目の牌の分のずれ
                        ),
                        pon[0][0],
                        rotate=True,
                        is_red=pon[0][1],
                        highliting=pon[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x,
                            770,
                        ),
                        pon[1][0],
                        is_red=pon[1][1],
                        highliting=pon[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin
                            + (len(hands[i]) + 1) * char_width
                            + left_x
                            + char_height
                            + char_width,
                            770,
                        ),
                        pon[2][0],
                        is_red=pon[2][1],
                        highliting=pon[0][2].is_highlighting and highlight_last_event,
                    )

                elif open_tile[1] == RelativePlayerIdx.RIGHT:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            741.3,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x
                            - char_width  # 1つ目の牌の分のずれ
                            - char_width,  # 2つ目の牌の分のずれ
                        ),
                        pon[0][0],
                        rotate=True,
                        is_red=pon[0][1],
                        highliting=pon[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x,
                            770,
                        ),
                        pon[1][0],
                        is_red=pon[1][1],
                        highliting=pon[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x + char_width,
                            770,
                        ),
                        pon[2][0],
                        is_red=pon[2][1],
                        highliting=pon[0][2].is_highlighting and highlight_last_event,
                    )
                left_x += char_width * 2 + char_height

            elif open_tile[2] == TileUnitType.CLOSED_KAN:
                closed_kan = open_tile[0]
                dwg_add(
                    dwg,
                    pai[i],
                    (
                        left_margin + (len(hands[i]) + 1) * char_width + left_x,
                        770,
                    ),
                    "\U0001F02B",
                    highliting=closed_kan[0][2].is_highlighting and highlight_last_event,
                )
                dwg_add(
                    dwg,
                    pai[i],
                    (
                        left_margin + (len(hands[i]) + 1) * char_width + left_x + char_width,
                        770,
                    ),
                    closed_kan[1][0],
                    is_red=closed_kan[1][1],
                    highliting=closed_kan[0][2].is_highlighting and highlight_last_event,
                )
                dwg_add(
                    dwg,
                    pai[i],
                    (
                        left_margin + (len(hands[i]) + 1) * char_width + left_x + char_width * 2,
                        770,
                    ),
                    closed_kan[2][0],
                    is_red=closed_kan[2][1],
                    highliting=closed_kan[0][2].is_highlighting and highlight_last_event,
                )
                dwg_add(
                    dwg,
                    pai[i],
                    (
                        left_margin + (len(hands[i]) + 1) * char_width + left_x + char_width * 3,
                        770,
                    ),
                    "\U0001F02B",
                    highliting=closed_kan[0][2].is_highlighting and highlight_last_event,
                )
                left_x += char_width * 4

            elif open_tile[2] == TileUnitType.OPEN_KAN:
                open_tile_kan = open_tile[0]
                if open_tile[1] == RelativePlayerIdx.LEFT:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            741.3,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x,
                        ),
                        open_tile_kan[0][0],
                        rotate=True,
                        is_red=open_tile_kan[0][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x + char_height,
                            770,
                        ),
                        open_tile_kan[1][0],
                        is_red=open_tile_kan[1][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin
                            + (len(hands[i]) + 1) * char_width
                            + left_x
                            + char_height
                            + char_width,
                            770,
                        ),
                        open_tile_kan[2][0],
                        is_red=open_tile_kan[2][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin
                            + (len(hands[i]) + 1) * char_width
                            + left_x
                            + char_height
                            + char_width * 2,
                            770,
                        ),
                        open_tile_kan[3][0],
                        is_red=open_tile_kan[3][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )

                elif open_tile[1] == RelativePlayerIdx.CENTER:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            741.3,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x
                            - char_width,
                        ),
                        txt=open_tile_kan[0][0],
                        rotate=True,
                        is_red=open_tile_kan[0][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x,
                            770,
                        ),
                        open_tile_kan[1][0],
                        is_red=open_tile_kan[1][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin
                            + (len(hands[i]) + 1) * char_width
                            + left_x
                            + char_height
                            + char_width,
                            770,
                        ),
                        open_tile_kan[2][0],
                        is_red=open_tile_kan[2][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin
                            + (len(hands[i]) + 1) * char_width
                            + left_x
                            + char_height
                            + char_width * 2,
                            770,
                        ),
                        open_tile_kan[3][0],
                        is_red=open_tile_kan[3][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )

                elif open_tile[1] == RelativePlayerIdx.RIGHT:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            741.3,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x
                            - char_width * 3,
                        ),
                        txt=open_tile_kan[0][0],
                        rotate=True,
                        is_red=open_tile_kan[0][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x,
                            770,
                        ),
                        open_tile_kan[1][0],
                        is_red=open_tile_kan[1][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x + char_width,
                            770,
                        ),
                        open_tile_kan[2][0],
                        is_red=open_tile_kan[2][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin
                            + (len(hands[i]) + 1) * char_width
                            + left_x
                            + char_height
                            + char_width * 2,
                            770,
                        ),
                        open_tile_kan[3][0],
                        is_red=open_tile_kan[3][1],
                        highliting=open_tile_kan[0][2].is_highlighting and highlight_last_event,
                    )
                left_x += char_width * 3 + char_height

            elif open_tile[2] == TileUnitType.ADDED_KAN:
                added_kan = open_tile[0]
                if open_tile[1] == RelativePlayerIdx.LEFT:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            741.3,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x,
                        ),
                        txt=added_kan[0][0],
                        rotate=True,
                        is_red=added_kan[0][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            710,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x,
                        ),
                        added_kan[1][0],
                        rotate=True,
                        is_red=added_kan[1][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x + char_height,
                            770,
                        ),
                        added_kan[2][0],
                        is_red=added_kan[2][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin
                            + (len(hands[i]) + 1) * char_width
                            + left_x
                            + char_height
                            + char_width,
                            770,
                        ),
                        added_kan[3][0],
                        is_red=added_kan[3][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )

                elif open_tile[1] == RelativePlayerIdx.CENTER:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            741.3,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x
                            - char_width,
                        ),
                        added_kan[1][0],
                        rotate=True,
                        is_red=added_kan[1][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            710,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x
                            - char_width,
                        ),
                        added_kan[2][0],
                        rotate=True,
                        is_red=added_kan[2][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x,
                            770,
                        ),
                        added_kan[0][0],
                        is_red=added_kan[0][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin
                            + (len(hands[i]) + 1) * char_width
                            + left_x
                            + char_height
                            + char_width,
                            770,
                        ),
                        added_kan[3][0],
                        is_red=added_kan[3][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )

                elif open_tile[1] == RelativePlayerIdx.RIGHT:
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            741.3,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x
                            - char_width * 2,
                        ),
                        added_kan[2][0],
                        rotate=True,
                        is_red=added_kan[2][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            710,
                            -left_margin  # 初期位置
                            - 3  # 回転時のずれ
                            - (len(hands[i]) + 1) * char_width  # 手牌の分のずれ
                            - left_x
                            - char_width * 2,
                        ),
                        added_kan[3][0],
                        rotate=True,
                        is_red=added_kan[3][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x,
                            770,
                        ),
                        added_kan[0][0],
                        is_red=added_kan[0][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                    dwg_add(
                        dwg,
                        pai[i],
                        (
                            left_margin + (len(hands[i]) + 1) * char_width + left_x + char_width,
                            770,
                        ),
                        added_kan[1][0],
                        is_red=added_kan[1][1],
                        highliting=added_kan[0][2].is_highlighting and highlight_last_event,
                    )
                left_x += char_width * 2 + char_height

        players[i].add(pai[i])
        players[i].add(player_info[i])
        players[i].rotate((360 - i * 90), (width / 2, height / 2))
        player_g.add(players[i])

    dwg.add(player_g)
    return dwg


def to_svg(json_data: str, target_idx: Optional[int] = None, show_name: bool = False) -> str:
    proto_data = MahjongTable.json_to_proto(json_data=json_data)
    dwg = _make_svg(proto_data, ".svg", target_idx, show_name=show_name)
    return dwg.tostring()


def save_svg(
    proto_data: Union[mjxproto.State, mjxproto.Observation],
    filename: str,
    target_idx: Optional[int] = None,
    highlight_last_event: bool = True,
) -> None:
    """Visualize State/Observation proto and save as svg file.

    Args
    ----
      proto_data: State or observation proto
      target_idx: the player you want to highlight
    """
    dwg = _make_svg(proto_data, filename, target_idx, highlight_last_event=highlight_last_event)
    dwg.save()



import json
import os
from typing import List

import mjxproto
from mjx.converter.mjlog_decoder import MjlogDecoder
from mjx.converter.mjscore_encoder import mjxproto_to_mjscore


def mjscore_log_equal(mjscore_original_dict, mjscore_converted_dict, proto) -> bool:
    original_log = mjscore_original_dict["log"][0]  # logのみを比べる
    converted_log = mjscore_converted_dict["log"][0]
    is_equal = True
    for i in range(len(original_log)):
        if original_log[i] != converted_log[i]:
            print(original_log[i], converted_log[i], proto.terminal.wins[0].ten_changes)
            is_equal = False
    return is_equal


def test_mjxproto_to_mjscore():
    mjlog_decoder = MjlogDecoder(modify=False)
    mjscore_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/mjscore")
    mjlog_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/mjlog")
    for mjscore in os.listdir(mjscore_dir):
        basename = os.path.splitext(mjscore)[0]
        mjscore_path = os.path.join(mjscore_dir, mjscore)
        mjlog_path = os.path.join(mjlog_dir, basename + ".mjlog")
        with open(mjscore_path, "r") as fs, open(mjlog_path, "r") as fl:
            lines = fl.readlines()
            assert len(lines) == 1
            mjlog = lines[0]
            mjxprotos: List[mjxproto.State] = mjlog_decoder.to_states(mjlog)
            mjscores = fs.readlines()
            assert len(mjxprotos) == len(mjscores)
            for proto, score_original in zip(mjxprotos, mjscores):
                score_converted = mjxproto_to_mjscore(proto)
                score_converted_dict = json.loads(score_converted)
                score_original_dict = json.loads(score_original)
                assert mjscore_log_equal(score_original_dict, score_converted_dict, proto)



from mjx.action import Action
from mjx.agent import Agent
from mjx.const import ActionType, EventType, TileType
from mjx.env import MjxEnv, run
from mjx.event import Event
from mjx.hand import Hand
from mjx.observation import Observation
from mjx.open import Open
from mjx.state import State
from mjx.tile import Tile

__all__ = [
    "Action",
    "Event",
    "Observation",
    "State",
    "MjxEnv",
    "Agent",
    "Open",
    "Hand",
    "Tile",
    "run",
    "ActionType",
    "EventType",
    "TileType",
]



# postpone type hint evaluation or doctest fails
from __future__ import annotations

import json
from typing import Dict, List

import mjxproto
from mjx.const import AbsolutePos, RelativePos
from mjx.converter import open_converter


def _change_tile_fmt(tile_id: int) -> int:
    reds_in_mjxproto = [16, 52, 88]
    reds_in_mjscore = [51, 52, 53]
    reds_dict = dict(zip(reds_in_mjxproto, reds_in_mjscore))
    if tile_id in reds_in_mjxproto:
        tile_in_mjscore = reds_dict[tile_id]
    else:
        tile_in_mjscore = ((tile_id // 36) + 1) * 10 + ((tile_id % 36) // 4) + 1
    return tile_in_mjscore


# mjxproto形式の牌のリストを引数にとり、表現をmjscore形式の表現に変える関数
def _change_tiles_fmt(tile_ids):
    # mjxproto形式の表現ををmjscore形式に変換
    scores = list(map(_change_tile_fmt, tile_ids))
    return scores


# openの情報を受け取ってmjscore形式に変更する関数
def _change_action_format(bits: int) -> str:  # TODO カン
    event_type = open_converter.open_event_type(bits)
    open_from = open_converter.open_from(bits)
    stolen_tile = open_converter.change_open_tile_fmt(open_converter.open_stolen_tile_type(bits))
    open_tiles = open_converter.change_open_tiles_fmt(open_converter.open_tile_types(bits))
    open_tiles.remove(stolen_tile)
    if event_type == mjxproto.EVENT_TYPE_CHI:  # チー
        return "c" + str(stolen_tile) + str(open_tiles[0]) + str(open_tiles[1])
    elif event_type == mjxproto.EVENT_TYPE_PON:  # ポン
        if open_from == RelativePos.LEFT:
            return "p" + str(stolen_tile) + str(open_tiles[0]) + str(open_tiles[1])
        elif open_from == RelativePos.MID:
            return str(open_tiles[0]) + "p" + str(stolen_tile) + str(open_tiles[1])
        else:
            return str(open_tiles[0]) + str(open_tiles[1]) + "p" + str(stolen_tile)
    elif event_type == mjxproto.EVENT_TYPE_ADDED_KAN:  # 加槓
        if open_from == RelativePos.LEFT:
            return (
                "k"
                + str(stolen_tile)
                + str(open_tiles[0])
                + str(open_tiles[1])
                + str(open_tiles[2])
            )
        elif open_from == RelativePos.MID:
            return (
                str(open_tiles[0])
                + "k"
                + str(stolen_tile)
                + str(open_tiles[1])
                + str(open_tiles[2])
            )
        else:
            return (
                str(open_tiles[0])
                + str(open_tiles[1])
                + "k"
                + str(stolen_tile)
                + str(open_tiles[2])
            )
    elif event_type == mjxproto.EVENT_TYPE_CLOSED_KAN:  # 暗槓
        return str(stolen_tile) + str(stolen_tile) + str(stolen_tile) + "a" + str(open_tiles[-1])
    else:  # 明槓
        if open_from == RelativePos.LEFT:
            return (
                "m"
                + str(stolen_tile)
                + str(open_tiles[0])
                + str(open_tiles[1])
                + str(open_tiles[2])
            )
        elif open_from == RelativePos.MID:
            return (
                str(open_tiles[0])
                + "m"
                + str(stolen_tile)
                + str(open_tiles[1])
                + str(open_tiles[2])
            )
        else:
            return (
                str(open_tiles[0])
                + str(open_tiles[1])
                + "m"
                + str(stolen_tile)
                + str(open_tiles[2])
            )


# mjscore形式の配牌をソートする関数。
def sort_init_hand(init_hand: List) -> List:
    # 11~19マンズ　21~29ピンズ　31~39ソウズ　#51~53赤マンピンソウ
    reds_score = [51, 52, 53]  # 赤
    init_key = [
        int(str(i)[::-1]) + 0.1 if i in reds_score else i for i in init_hand
    ]  # ソートする辞書のキー。赤は文字を反転させて、同じ種類の牌の中で最後尾になるようにソートする値に0.1を足す。
    init_hand_dict = [[k, v] for k, v in zip(init_key, init_hand)]
    sorted_hand = sorted(init_hand_dict, key=lambda x: x[0])
    sorted_hand = [i[1] for i in sorted_hand]
    return sorted_hand


def _change_tumogiri_riich_fmt(tile):  # ツモギリリーチ専用の番号90を60ツモぎりの番号60に直す
    if tile == 90:
        return 60
    return tile


# mjxproto形式のeventを受け取り、あるプレイヤーの捨て牌をmjscore形式で出力する関数。
def parse_discards(events, abs_pos: int):
    discards: List[object] = []
    for i, event in enumerate(events):
        if event.type == mjxproto.EVENT_TYPE_DISCARD and event.who == abs_pos:  # 手出し
            if events[i - 1].type == mjxproto.EVENT_TYPE_RIICHI:  # 一つ前のeventがriichiかどうか
                discards.append("r" + str(_change_tile_fmt(event.tile)))
            else:
                discards.append(_change_tile_fmt(event.tile))
        elif event.type == mjxproto.EVENT_TYPE_TSUMOGIRI and event.who == abs_pos:  # ツモギリ
            if events[i - 1].type == mjxproto.EVENT_TYPE_RIICHI:  # 一つ前のeventがriichiかどうか
                discards.append("r60")
            else:
                discards.append(60)
        elif event.type == mjxproto.EVENT_TYPE_CLOSED_KAN and event.who == abs_pos:
            discards.append(_change_action_format(event.open))
        elif event.type == mjxproto.EVENT_TYPE_ADDED_KAN and event.who == abs_pos:
            discards.append(_change_action_format(event.open))
        elif (
            events[i - 1].type == mjxproto.EVENT_TYPE_OPEN_KAN and event.who == abs_pos
        ):  # 明槓のあと捨て牌の情報に情報のない0が追加される。
            discards.append(0)
    return discards


# プレイヤーごとにmjscore形式のdraw_historyを返す。
def parse_draw_history(draw_history, events, abs_pos):
    """
    - mjscoreでは引いた牌のリストにチーやポンなどのアクションが含まれている
    - mjxprotoの　draw_historyでは単に飛ばされていて、eventの方に情報がある

    方針
    1. チーポンも含めたdiscardsを作成
    2. draw_historyの方で直前の捨て牌の直後にアクションを挿入
    """
    draw_history = _change_tiles_fmt(draw_history)
    discards = []
    actions = []  #
    for i, event in enumerate(events):
        if event.type == mjxproto.EVENT_TYPE_DISCARD and event.who == abs_pos:  # 手出し
            discards.append(event.tile)
        elif event.type == mjxproto.EVENT_TYPE_TSUMOGIRI and event.who == abs_pos:  # ツモギリ
            discards.append(60)
        elif event.type == mjxproto.EVENT_TYPE_CHI and event.who == abs_pos:  # チー
            discards.append(event.open)
            actions.append(event.open)
        elif event.type == mjxproto.EVENT_TYPE_PON and event.who == abs_pos:  # ポン
            discards.append(event.open)
            actions.append(event.open)
        elif event.type == mjxproto.EVENT_TYPE_OPEN_KAN and event.who == abs_pos:  # 明槓
            discards.append(event.open)
            actions.append(event.open)
        elif (
            event.type == mjxproto.EVENT_TYPE_CLOSED_KAN and event.who == abs_pos
        ):  # 捨て牌の情報には暗槓も含まれているので、追加しないとずれる。
            discards.append(_change_action_format(event.open))
    for i, action in enumerate(actions):
        # 捨て牌でのactionのindex同じ順にdrawにアクションを挿入すれば良い
        action_index = discards.index(action) - i
        draw_history.insert(action_index, _change_action_format(action))
    return draw_history


yaku_list = [
    "門前清自摸和",
    "立直",
    "一発",
    "槍槓",
    "嶺上開花",
    "海底摸月",
    "河底撈魚",
    "平和",
    "断幺九",
    "一盃口",
    "自風 東",
    "自風 南",
    "自風 西",
    "自風 北",
    "場風 東",
    "場風 南",
    "場風 西",
    "場風 北",
    "役牌 白",
    "役牌 發",
    "役牌 中",
    "両立直",
    "七対子",
    "混全帯幺九",
    "一気通貫",
    "三色同順",
    "三色同刻",
    "三槓子",
    "対々和",
    "三暗刻",
    "小三元",
    "混老頭",
    "二盃口",
    "純全帯幺九",
    "混一色",
    "清一色",
    "人和",  # 天鳳は人和なし
    "天和(役満)",
    "地和(役満)",
    "大三元(役満)",
    "四暗刻(役満)",
    "四暗刻単騎(役満)",
    "字一色(役満)",
    "緑一色(役満)",
    "清老頭(役満)",
    "九蓮宝燈(役満)",
    "純正九蓮宝燈(役満)",
    "国士無双(役満)",
    "国士無双１３面(役満)",
    "大四喜(役満)",
    "小四喜(役満)",
    "四槓子(役満)",
    "ドラ",
    "裏ドラ",
    "赤ドラ",
]


yaku_list_keys = [i for i in range(55)]
yaku_dict = {k: v for k, v in zip(yaku_list_keys, yaku_list)}

non_dealer_tsumo_dict = {
    1100: "300-500",
    1500: "400-700",
    1600: "800-1600",
    2000: "500-1000",
    2400: "600-1200",
    2700: "700-1300",
    3100: "800-1500",
    3200: "800-1600",
    3600: "900-1800",
    4000: "1000-2000",
    4700: "1200-2300",
    5200: "1300-2600",
    5900: "1500-2900",
    6400: "1600-3200",
    7200: "1800-3600",
    7900: "2000-3900",
    8000: "2000-4000",
    12000: "3000-6000",
    16000: "4000-8000",
    24000: "6000-12000",
    32000: "8000-16000",
}
no_winner_dict = {
    mjxproto.EVENT_TYPE_EXHAUSTIVE_DRAW_NORMAL: "流局",
    mjxproto.EVENT_TYPE_ABORTIVE_DRAW_NINE_TERMINALS: "九種九牌",
    mjxproto.EVENT_TYPE_ABORTIVE_DRAW_FOUR_RIICHIS: "四家立直",
    mjxproto.EVENT_TYPE_ABORTIVE_DRAW_THREE_RONS: "三家和了",
    mjxproto.EVENT_TYPE_ABORTIVE_DRAW_FOUR_KANS: "四槓散了",
    mjxproto.EVENT_TYPE_ABORTIVE_DRAW_FOUR_WINDS: "四風連打",
}
dealer_point_dict = {12000: "満貫", 18000: "跳満", 24000: "倍満", 36000: "三倍満", 48000: "役満"}
no_dealer_point_dict = {8000: "満貫", 12000: "跳満", 16000: "倍満", 24000: "三倍満", 32000: "役満"}


def _fan_fu(who, fans: List[int], fu: int, ten, round: int) -> str:
    """
    >>> _fan_fu(0, [3, 1], 40, 12000, 0)
    '満貫'
    >>> _fan_fu(1, [2, 1], 40, 5200, 0)
    '40符3飜'
    """
    fan = sum(fans)
    if _is_oya(who, round):  # 親かどうかを判定
        if ten < 12000:
            return str(fu) + "符" + str(fan) + "飜"
        else:
            return dealer_point_dict[ten]
    else:
        if int(ten) < 8000:
            return str(fu) + "符" + str(fan) + "飜"
        else:
            return no_dealer_point_dict[ten]


def _is_oya(who: int, round: int) -> bool:  # 和了者が親かどうか判定する。
    """
    >>> _is_oya(0, 3)
    False
    >>> _is_oya(0, 4)
    True
    """
    if round % 4 == who:
        return True
    else:
        return False


def _winner_point(who: int, from_who: int, fans: List[int], fu: int, ten: int, round: int) -> str:
    """
    >>> _winner_point(0, 0, [3, 0], 30, 6000, 0)
    '30符3飜2000点∀'
    >>> _winner_point(2, 0, [5, 0], 40, 8000, 0)
    '満貫8000点'
    """
    is_tsumo = who == from_who  # ツモあがりかどうかを判定
    if is_tsumo:
        if _is_oya(who, round):  # 親かどうか
            return _fan_fu(who, fans, fu, ten, round) + str(int(ten / 3)) + "点∀"
        else:
            return _fan_fu(who, fans, fu, ten, round) + non_dealer_tsumo_dict[ten] + "点"
    else:
        if who == AbsolutePos.INIT_EAST:
            return _fan_fu(who, fans, fu, ten, round) + str(ten) + "点"
        else:
            return _fan_fu(who, fans, fu, ten, round) + str(ten) + "点"


def _ditermin_yaku_list(
    fans: List[int], yakus: List[int], yakumans: List[int]
) -> List[int]:  # リーチがかかるとprotoではyakus
    # に強制的にウラドラの情報が入るが、乗っているかどうかを確認する必要がある
    """
    >>> _ditermin_yaku_list([1, 1, 1, 0], [1, 0, 7, 53], [])
    [1, 0, 7]
    >>> _ditermin_yaku_list([1, 1, 2, 1, 2, 0], [1, 0, 29, 8, 54, 53], [])
    [1, 0, 29, 8, 54]
    """
    if len(yakumans) != 0:
        return yakumans
    if sum(fans) < len(yakus):
        return [i for i in yakus if i != 53]
    elif fans[-1] == 0:  # 裏ドラは必ずfansの末尾に表示されるので0かどうかで判定がつく。
        return [i for i in yakus if i != 53]
    else:
        return yakus


def _correspond_yakus(yaku_dict, yakus: List[int], fans: List[int]):
    """
    >>> _correspond_yakus(yaku_dict, [0, 52], [1, 2])
    ['門前清自摸和(1飜)', 'ドラ(2飜)']
    """
    yakus_in_japanese = []
    for idx, yaku in enumerate(yakus):
        yakus_in_japanese.append(
            yaku_dict[yaku] + "({}飜)".format(str(fans[idx]))
        )  # ドラは複数ある場合はまとめてドラ(3飜)の様に表記
    return yakus_in_japanese


def _winner_yakus(yakus: List[int], fans: List[int], yakumans: List[int]) -> List[str]:
    """
    >>> _winner_yakus([0, 1, 23], [1, 1, 2], [])
    ['門前清自摸和(1飜)', '立直(1飜)', '混全帯幺九(2飜)']
    >>> _winner_yakus([23], [1], [])
    ['混全帯幺九(1飜)']
    """
    if len(yakumans) != 0:
        return [yaku_dict[i] for i in yakumans]
    return _correspond_yakus(yaku_dict, yakus, fans)


def _yaku_point_info(state: mjxproto.State, winner_num: int):
    round = state.public_observation.init_score.round
    who = state.round_terminal.wins[winner_num].who
    from_who = state.round_terminal.wins[winner_num].from_who
    fans = [i for i in state.round_terminal.wins[winner_num].fans]  # [役での飜数, ドラの数]
    yakumans = [i for i in state.round_terminal.wins[winner_num].yakumans]
    yakus = _ditermin_yaku_list(
        fans, [i for i in state.round_terminal.wins[winner_num].yakus], yakumans
    )
    fu = state.round_terminal.wins[winner_num].fu
    ten = state.round_terminal.wins[winner_num].ten
    # fnas:[役での飜数, ドラでの飜数]
    # yakus: [役とドラの種類]
    # ten: 純粋に上がり点が表示される。ツモ上がりの際の対応が必要
    yaku_point_infos = [who, from_who, who, _winner_point(who, from_who, fans, fu, ten, round)]
    yaku_point_infos.extend(_winner_yakus(yakus, fans, yakumans))
    return yaku_point_infos


def parse_terminal(state: mjxproto.State):
    if len(state.round_terminal.wins) == 0:  # あがった人がいない場合,# state.terminal.winsの長さは0
        ten_changes = [i for i in state.round_terminal.no_winner.ten_changes]
        if state.public_observation.events[-1].type == mjxproto.EVENT_TYPE_EXHAUSTIVE_DRAW_NORMAL:
            if len(state.round_terminal.no_winner.tenpais) == 0:
                return ["全員不聴"]
            else:
                return ["流局", ten_changes]
        if (
            state.public_observation.events[-1].type
            == mjxproto.EVENT_TYPE_EXHAUSTIVE_DRAW_NAGASHI_MANGAN
        ):
            # 流し満貫はten_changes も表示される。
            return ["流し満貫", ten_changes]
        return [no_winner_dict[state.public_observation.events[-1].type]]
    else:
        terminal_info: List = ["和了"]
        for i in range(len(state.round_terminal.wins)):  # ダブロンに対応するために上がり者の数に応じてfor文を回すようにする。
            ten_changes = [i for i in state.round_terminal.wins[i].ten_changes]
            yaku_point_info = _yaku_point_info(state, i)
            terminal_info.append(ten_changes)
            terminal_info.append(yaku_point_info)
        return terminal_info


def determine_ura_doras_list(state: mjxproto.State) -> List:
    if len(state.round_terminal.wins) == 0:  # あがり者の有無でウラどらが表示されるかどうかが決まる
        return []
    has_riichi = (
        1 not in state.round_terminal.wins[0].yakus
        and 21 not in state.round_terminal.wins[0].yakus
    )
    if has_riichi:  # リーチまたはダブリーがかかっていないと、上がって裏ドラが表示されない.
        return []
    return [_change_tile_fmt(i) for i in state.hidden_state.ura_dora_indicators]


# ここを実装
def mjxproto_to_mjscore(state: mjxproto.State) -> str:
    round: int = state.public_observation.init_score.round
    honba: int = state.public_observation.init_score.honba
    riichi: int = state.public_observation.init_score.riichi
    doras: List[int] = [_change_tile_fmt(i) for i in state.public_observation.dora_indicators]
    ura_doras = determine_ura_doras_list(state)
    init_score: List[int] = [i for i in state.public_observation.init_score.tens]
    log = [[round, honba, riichi], init_score, doras, ura_doras]
    absolute_pos = [
        AbsolutePos.INIT_EAST,
        AbsolutePos.INIT_SOUTH,
        AbsolutePos.INIT_WEST,
        AbsolutePos.INIT_NORTH,
    ]
    for abs_pos in absolute_pos:
        log.append(
            sort_init_hand(
                _change_tiles_fmt(state.private_observations[abs_pos].init_hand.closed_tiles)
            )
        )
        log.append(
            parse_draw_history(
                state.private_observations[abs_pos].draw_history,
                state.public_observation.events,
                abs_pos,
            )
        )
        log.append(parse_discards(state.public_observation.events, abs_pos))

    log.append(parse_terminal(state))
    d: Dict = {"title": [], "name": [], "rule": [], "log": [log]}
    return json.dumps(d)



from enum import IntEnum

import mjxproto


class ActionType(IntEnum):
    # After draw
    DISCARD = mjxproto.ACTION_TYPE_DISCARD
    TSUMOGIRI = mjxproto.ACTION_TYPE_TSUMOGIRI
    RIICHI = mjxproto.ACTION_TYPE_RIICHI
    CLOSED_KAN = mjxproto.ACTION_TYPE_CLOSED_KAN
    ADDED_KAN = mjxproto.ACTION_TYPE_ADDED_KAN
    TSUMO = mjxproto.ACTION_TYPE_TSUMO
    ABORTIVE_DRAW_NINE_TERMINALS = mjxproto.ACTION_TYPE_ABORTIVE_DRAW_NINE_TERMINALS  # 九種九牌
    # After other's discard
    CHI = mjxproto.ACTION_TYPE_CHI
    PON = mjxproto.ACTION_TYPE_PON
    OPEN_KAN = mjxproto.ACTION_TYPE_OPEN_KAN
    RON = mjxproto.ACTION_TYPE_RON
    PASS = mjxproto.ACTION_TYPE_NO
    # Dummy used only to check connection and share round terminal information
    DUMMY = mjxproto.ACTION_TYPE_DUMMY


class EventType(IntEnum):
    # Publicly observable actions
    DISCARD = mjxproto.EVENT_TYPE_DISCARD
    TSUMOGIRI = mjxproto.EVENT_TYPE_TSUMOGIRI  # ツモ切り, Tsumogiri
    RIICHI = mjxproto.EVENT_TYPE_RIICHI
    CLOSED_KAN = mjxproto.EVENT_TYPE_CLOSED_KAN
    ADDED_KAN = mjxproto.EVENT_TYPE_ADDED_KAN
    TSUMO = mjxproto.EVENT_TYPE_TSUMO
    ABORTIVE_DRAW_NINE_TERMINALS = mjxproto.EVENT_TYPE_ABORTIVE_DRAW_NINE_TERMINALS
    CHI = mjxproto.EVENT_TYPE_CHI
    PON = mjxproto.EVENT_TYPE_PON
    OPEN_KAN = mjxproto.EVENT_TYPE_OPEN_KAN
    RON = mjxproto.EVENT_TYPE_RON
    # State transitions made by environment. There is no decision making by players.
    # 11 is skipped for the consistency to ActionType
    DRAW = mjxproto.EVENT_TYPE_DRAW
    RIICHI_SCORE_CHANGE = mjxproto.EVENT_TYPE_RIICHI_SCORE_CHANGE
    NEW_DORA = mjxproto.EVENT_TYPE_NEW_DORA
    ABORTIVE_DRAW_FOUR_RIICHIS = mjxproto.EVENT_TYPE_ABORTIVE_DRAW_FOUR_RIICHIS  # 四家立直
    ABORTIVE_DRAW_THREE_RONS = mjxproto.EVENT_TYPE_ABORTIVE_DRAW_THREE_RONS  # 三家和了
    ABORTIVE_DRAW_FOUR_KANS = mjxproto.EVENT_TYPE_ABORTIVE_DRAW_FOUR_KANS  # 四槓散了
    ABORTIVE_DRAW_FOUR_WINDS = mjxproto.EVENT_TYPE_ABORTIVE_DRAW_FOUR_WINDS  # 四風連打
    ABORTIVE_DRAW_NORMAL = mjxproto.EVENT_TYPE_EXHAUSTIVE_DRAW_NORMAL  # 通常流局
    ABORTIVE_DRAW_NAGASHI_MANGAN = mjxproto.EVENT_TYPE_EXHAUSTIVE_DRAW_NAGASHI_MANGAN  # 流し満貫


class AbsolutePos(IntEnum):
    INIT_EAST = 0
    INIT_SOUTH = 1
    INIT_WEST = 2
    INIT_NORTH = 3


class RelativePos(IntEnum):
    SELF = 0
    RIGHT = 1
    MID = 2
    LEFT = 3


class SeedType(IntEnum):
    RANDOM = 0
    DUPLICATE = 1


class TileType(IntEnum):
    M1 = 0
    M2 = 1
    M3 = 2
    M4 = 3
    M5 = 4
    M6 = 5
    M7 = 6
    M8 = 7
    M9 = 8
    P1 = 9
    P2 = 10
    P3 = 11
    P4 = 12
    P5 = 13
    P6 = 14
    P7 = 15
    P8 = 16
    P9 = 17
    S1 = 18
    S2 = 19
    S3 = 20
    S4 = 21
    S5 = 22
    S6 = 23
    S7 = 24
    S8 = 25
    S9 = 26
    EW = 27
    SW = 28
    WW = 29
    NW = 30
    WD = 31
    GD = 32
    RD = 33


class RelativePlayerIdx(IntEnum):
    SELF = 0  # 自家
    RIGHT = 1  # 下家
    CENTER = 2  # 対面
    LEFT = 3  # 上家



from __future__ import annotations

from typing import List, Optional

import _mjx  # type: ignore
import numpy as np  # type: ignore
from google.protobuf import json_format

import mjxproto
from mjx.action import Action
from mjx.const import TileType
from mjx.event import Event
from mjx.hand import Hand
from mjx.tile import Tile
from mjx.visualizer.svg import save_svg


class Observation:
    def __init__(self, obs_json: Optional[str] = None) -> None:
        self._cpp_obj: Optional[_mjx.Observation] = None  # type: ignore
        if obs_json is None:
            return

        self._cpp_obj = _mjx.Observation(obs_json)  # type: ignore

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Observation):
            raise NotImplementedError
        raise NotImplementedError  # TODO: implement

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Observation):
            raise NotImplementedError
        raise NotImplementedError  # TODO: implement

    def who(self) -> int:
        assert self._cpp_obj is not None
        return self._cpp_obj.who()  # type: ignore

    def dealer(self) -> int:
        assert self._cpp_obj is not None
        return self._cpp_obj.dealer()  # type: ignore

    def doras(self) -> List[TileType]:
        assert self._cpp_obj is not None
        return self._cpp_obj.doras()  # type: ignore

    def curr_hand(self) -> Hand:
        assert self._cpp_obj is not None
        return Hand._from_cpp_obj(self._cpp_obj.curr_hand())  # type: ignore

    def legal_actions(self) -> List[Action]:
        assert self._cpp_obj is not None
        return [Action._from_cpp_obj(cpp_obj) for cpp_obj in self._cpp_obj.legal_actions()]  # type: ignore

    def events(self) -> List[Event]:
        assert self._cpp_obj is not None
        return [Event._from_cpp_obj(cpp_obj) for cpp_obj in self._cpp_obj.events()]  # type: ignore

    def draw_history(self) -> List[Tile]:
        assert self._cpp_obj is not None
        return [Tile(t) for t in self._cpp_obj.draw_history()]  # type: ignore

    def action_mask(self, dtype=np.float32) -> np.ndarray:
        assert self._cpp_obj is not None
        return np.array(self._cpp_obj.action_mask(), dtype=dtype)  # type: ignore

    def kyotaku(self) -> int:
        assert self._cpp_obj is not None
        return self._cpp_obj.kyotaku()

    def honba(self) -> int:
        assert self._cpp_obj is not None
        return self._cpp_obj.honba()

    def tens(self) -> List[int]:
        assert self._cpp_obj is not None
        return self._cpp_obj.tens()

    def rankings(self) -> List[int]:
        assert self._cpp_obj is not None
        tens = self.tens()
        order = list(range(4))
        order.sort(key=lambda i: tens[i] - i)
        order.reverse()
        ret = [0] * 4
        for i, p in enumerate(order):
            ret[p] = i

        # e.g tens = [20000, 35000, 35000, 10000]
        #     => order = [1, 2, 0, 3]
        #        ret   = [2, 0, 1, 3]

        return ret

    def round(self) -> int:
        # 東一局:0, ..., 南四局:7, ...
        assert self._cpp_obj is not None
        return self._cpp_obj.round()

    def to_json(self) -> str:
        assert self._cpp_obj is not None
        return self._cpp_obj.to_json()  # type: ignore

    def to_proto(self) -> mjxproto.Observation:
        assert self._cpp_obj is not None
        return json_format.Parse(self.to_json(), mjxproto.Observation())

    def save_svg(self, filename: str, view_idx: Optional[int] = None) -> None:
        assert filename.endswith(".svg")
        assert view_idx is None or 0 <= view_idx < 4

        observation = self.to_proto()
        save_svg(observation, filename, view_idx)

    @staticmethod
    def add_legal_actions(obs_json: str) -> str:
        assert len(Observation(obs_json).legal_actions()) == 0, "Legal actions are alredy set."
        return _mjx.Observation.add_legal_actions(obs_json)

    @classmethod
    def from_proto(cls, proto: mjxproto.Observation) -> Observation:
        return Observation(json_format.MessageToJson(proto))

    @classmethod
    def _from_cpp_obj(cls, cpp_obj) -> Observation:
        obs = cls()
        obs._cpp_obj = cpp_obj
        return obs



from typing import List

import mjxproto
from mjx.converter.open_converter import open_event_type

MASK_CHI_OFFSET = [0b0000000000011000, 0b0000000001100000, 0b0000000110000000]
MASK_PON_UNUSED_OFFSET = 0b0000000001100000


def open_tile_ids(bits: int) -> List[int]:
    """
    各鳴きで使われた牌のidを返す関数
    以下、各鳴きに対するdoctest
    >>> open_tile_ids(49495)  # chi
    [82, 86, 90]
    >>> open_tile_ids(47723)  # pon
    [124, 125, 126]
    >>> open_tile_ids(1793)  # open_kan
    [4, 5, 6, 7]
    >>> open_tile_ids(2048)  # closed_kan
    [8, 9, 10, 11]
    >>> open_tile_ids(530)  # added_kan
    [0, 1, 2, 3]
    """
    event_type = open_event_type(bits)
    if event_type == mjxproto.EVENT_TYPE_CHI:
        return chi_ids_(bits)
    elif event_type == mjxproto.EVENT_TYPE_PON:
        return pon_ids_(bits)
    elif event_type == mjxproto.EVENT_TYPE_OPEN_KAN:
        return oepn_kan_ids_(bits)
    elif event_type == mjxproto.EVENT_TYPE_CLOSED_KAN:
        return closed_kan_ids_(bits)
    else:
        return added_kan_ids_(bits)


def chi_ids_(bits: int) -> List[int]:
    """
    チーに使われた牌のidを返す関数
    """
    min_type_base = (bits >> 10) // 3
    min_type = (min_type_base // 7) * 9 + min_type_base % 7
    return [(min_type + i) * 4 + ((bits & MASK_CHI_OFFSET[i]) >> (2 * i + 3)) for i in range(3)]


def pon_ids_(bits: int) -> List[int]:
    """
    ポンに使われた牌のidを返す関数
    """
    ids = []
    for i in range(3):
        type = (bits >> 9) // 3
        unused_offset = (bits & MASK_PON_UNUSED_OFFSET) >> 5
        if i >= unused_offset:
            i = i + 1
        ids.append(type * 4 + i)
    return ids


def oepn_kan_ids_(bits: int) -> List[int]:
    """
    大明槓に使われた牌のidを返す関数
    """
    return [((bits >> 8) // 4) * 4 + i for i in range(4)]


def closed_kan_ids_(bits: int) -> List[int]:
    """
    暗槓に使われた牌のidを返す関数
    """
    return [((bits >> 8) // 4) * 4 + i for i in range(4)]


def added_kan_ids_(bits: int) -> List[int]:
    """
    加槓に使われた牌のidを返す関数
    """
    return [((bits >> 9) // 3) * 4 + i for i in range(4)]



from google.protobuf import json_format

import mjx
import mjxproto
from mjx.const import ActionType

json_str = '{"tile":10}'


def test_Action():
    action = mjx.Action(json_str)
    restored = action.to_json()
    assert json_str == restored


def test_to_json():
    action = mjx.Action(json_str)
    restored = action.to_json()
    assert json_str == restored


def test_to_proto():
    action = mjx.Action(json_str)
    action.to_proto() == json_format.Parse(json_str, mjxproto.Action())


def test_to_idx():
    action = mjx.Action(json_str)
    idx = action.to_idx()
    assert idx == 2


def test_select_from():
    legal_actions = [
        mjx.Action('{"tile":10}'),
        mjx.Action('{"tile":1}'),
    ]
    action = mjx.Action.select_from(0, legal_actions)
    assert action.to_json() == '{"tile":1}'


def test_from_proto():
    action = mjx.Action(json_str)
    proto_action = action.to_proto()
    mjx.Action.from_proto(proto_action).to_json() == json_str


def test_type():
    action = mjx.Action(json_str)
    assert action.type() == 0
    assert action.type() == ActionType.DISCARD
    assert action.type() != 1
    assert action.type() != ActionType.TSUMOGIRI


def test_tile():
    action = mjx.Action('{"tile":10}')
    tile = action.tile()
    assert tile is not None
    assert tile.id() == 10

    action = mjx.Action('{"who":1,"type":"ACTION_TYPE_CHI","open":30111}')
    tile = action.tile()
    assert tile is None



from __future__ import annotations

from typing import Optional

from mjx.const import EventType, RelativePlayerIdx
from mjxproto import ActionType

to_char = [
    "m1",
    "m2",
    "m3",
    "m4",
    "m5",
    "m6",
    "m7",
    "m8",
    "m9",
    "p1",
    "p2",
    "p3",
    "p4",
    "p5",
    "p6",
    "p7",
    "p8",
    "p9",
    "s1",
    "s2",
    "s3",
    "s4",
    "s5",
    "s6",
    "s7",
    "s8",
    "s9",
    "ew",
    "sw",
    "ww",
    "nw",
    "wd",
    "gd",
    "rd",
]

to_unicode = [
    "\U0001F007",
    "\U0001F008",
    "\U0001F009",
    "\U0001F00A",
    "\U0001F00B",
    "\U0001F00C",
    "\U0001F00D",
    "\U0001F00E",
    "\U0001F00F",
    "\U0001F019",
    "\U0001F01A",
    "\U0001F01B",
    "\U0001F01C",
    "\U0001F01D",
    "\U0001F01E",
    "\U0001F01F",
    "\U0001F020",
    "\U0001F021",
    "\U0001F010",
    "\U0001F011",
    "\U0001F012",
    "\U0001F013",
    "\U0001F014",
    "\U0001F015",
    "\U0001F016",
    "\U0001F017",
    "\U0001F018",
    "\U0001F000",
    "\U0001F001",
    "\U0001F002",
    "\U0001F003",
    "\U0001F006",
    "\U0001F005",
    "\U0001F004\uFE0E",
]

to_wind_char = [
    "EAST",
    "SOUTH",
    "WEST",
    "NORTH",
    "東",
    "南",
    "西",
    "北",
]

to_relative_player_idx = {
    RelativePlayerIdx.RIGHT: "R",  # Right
    RelativePlayerIdx.CENTER: "C",  # Center
    RelativePlayerIdx.LEFT: "L",  # Left
    RelativePlayerIdx.SELF: "S",  # Self(kan closed)
}

yaku_list = [
    "門前清自摸和",
    "立直",
    "一発",
    "槍槓",
    "嶺上開花",
    "海底摸月",
    "河底撈魚",
    "平和",
    "断幺九",
    "一盃口",
    "自風 東",
    "自風 南",
    "自風 西",
    "自風 北",
    "場風 東",
    "場風 南",
    "場風 西",
    "場風 北",
    "役牌 白",
    "役牌 發",
    "役牌 中",
    "両立直",
    "七対子",
    "混全帯幺九",
    "一気通貫",
    "三色同順",
    "三色同刻",
    "三槓子",
    "対々和",
    "三暗刻",
    "小三元",
    "混老頭",
    "二盃口",
    "純全帯幺九",
    "混一色",
    "清一色",
    "人和",  # 天鳳は人和なし
    "天和(役満)",
    "地和(役満)",
    "大三元(役満)",
    "四暗刻(役満)",
    "四暗刻単騎(役満)",
    "字一色(役満)",
    "緑一色(役満)",
    "清老頭(役満)",
    "九蓮宝燈(役満)",
    "純正九蓮宝燈(役満)",
    "国士無双(役満)",
    "国士無双１３面(役満)",
    "大四喜(役満)",
    "小四喜(役満)",
    "四槓子(役満)",
    "ドラ",
    "裏ドラ",
    "赤ドラ",
]

event_type_en = {
    EventType.DISCARD: "DISCARD",
    EventType.TSUMOGIRI: "TSUMOGIRI",
    EventType.RIICHI: "RIICHI",
    EventType.CLOSED_KAN: "CLOSED_KAN",
    EventType.ADDED_KAN: "ADDED_KAN",
    EventType.TSUMO: "TSUMO",
    EventType.ABORTIVE_DRAW_NINE_TERMINALS: "ABORTIVE_DRAW_NINE_TERMINALS",
    EventType.CHI: "CHI",
    EventType.PON: "PON",
    EventType.OPEN_KAN: "OPEN_KAN",
    EventType.RON: "RON",
    EventType.DRAW: "DRAW",
    EventType.RIICHI_SCORE_CHANGE: "RIICHI_SCORE_CHANGE",
    EventType.NEW_DORA: "NEW_DORA",
    EventType.ABORTIVE_DRAW_FOUR_RIICHIS: "ABORTIVE_DRAW_FOUR_RIICHIS",
    EventType.ABORTIVE_DRAW_THREE_RONS: "ABORTIVE_DRAW_THREE_RONS",
    EventType.ABORTIVE_DRAW_FOUR_KANS: "ABORTIVE_DRAW_FOUR_KANS",
    EventType.ABORTIVE_DRAW_FOUR_WINDS: "ABORTIVE_DRAW_FOUR_WINDS",
    EventType.ABORTIVE_DRAW_NORMAL: "ABORTIVE_DRAW_NORMAL",
    EventType.ABORTIVE_DRAW_NAGASHI_MANGAN: "ABORTIVE_DRAW_NAGASHI_MANGAN",
}

event_type_ja = {
    EventType.DISCARD: "打牌",
    EventType.TSUMOGIRI: "ツモ切り",
    EventType.RIICHI: "リーチ",
    EventType.CLOSED_KAN: "暗槓",
    EventType.ADDED_KAN: "加槓",
    EventType.TSUMO: "ツモ",
    EventType.ABORTIVE_DRAW_NINE_TERMINALS: "九種九牌",
    EventType.CHI: "チー",
    EventType.PON: "ポン",
    EventType.OPEN_KAN: "大明槓",
    EventType.RON: "ロン",
    EventType.DRAW: "引き分け",
    EventType.RIICHI_SCORE_CHANGE: "リーチ（スコア変動）",
    EventType.NEW_DORA: "新ドラ",
    EventType.ABORTIVE_DRAW_FOUR_RIICHIS: "四家立直",
    EventType.ABORTIVE_DRAW_THREE_RONS: "三家和了",
    EventType.ABORTIVE_DRAW_FOUR_KANS: "四槓散了",
    EventType.ABORTIVE_DRAW_FOUR_WINDS: "四風連打",
    EventType.ABORTIVE_DRAW_NORMAL: "流局",
    EventType.ABORTIVE_DRAW_NAGASHI_MANGAN: "流し満貫",
}

action_type_en = {
    ActionType.ACTION_TYPE_DISCARD: "DISCARD",
    ActionType.ACTION_TYPE_TSUMOGIRI: "TSUMOGIRI",
    ActionType.ACTION_TYPE_RIICHI: "RIICHI",
    ActionType.ACTION_TYPE_CLOSED_KAN: "CLOSED_KAN",
    ActionType.ACTION_TYPE_ADDED_KAN: "ADDED_KAN",
    ActionType.ACTION_TYPE_TSUMO: "TSUMO",
    ActionType.ACTION_TYPE_ABORTIVE_DRAW_NINE_TERMINALS: "ABORTIVE_DRAW_NINE_TERMINALS",
    ActionType.ACTION_TYPE_CHI: "CHI",
    ActionType.ACTION_TYPE_PON: "PON",
    ActionType.ACTION_TYPE_OPEN_KAN: "OPEN_KAN",
    ActionType.ACTION_TYPE_RON: "RON",
    ActionType.ACTION_TYPE_NO: "PASS",
    ActionType.ACTION_TYPE_DUMMY: "DUMMY",
}

action_type_ja = {
    ActionType.ACTION_TYPE_DISCARD: "打牌",
    ActionType.ACTION_TYPE_TSUMOGIRI: "ツモ切り",
    ActionType.ACTION_TYPE_RIICHI: "リーチ",
    ActionType.ACTION_TYPE_CLOSED_KAN: "暗槓",
    ActionType.ACTION_TYPE_ADDED_KAN: "加槓",
    ActionType.ACTION_TYPE_TSUMO: "ツモ",
    ActionType.ACTION_TYPE_ABORTIVE_DRAW_NINE_TERMINALS: "九種九牌",
    ActionType.ACTION_TYPE_CHI: "チー",
    ActionType.ACTION_TYPE_PON: "ポン",
    ActionType.ACTION_TYPE_OPEN_KAN: "明槓",
    ActionType.ACTION_TYPE_RON: "ロン",
    ActionType.ACTION_TYPE_NO: "パス",
    ActionType.ACTION_TYPE_DUMMY: "ダミー",
}


def get_tile_char(tile_id: int, is_using_unicode: bool) -> str:
    if tile_id < 0 or 136 < tile_id:
        return " "
    if is_using_unicode:
        return to_unicode[tile_id // 4]
    else:
        return to_char[tile_id // 4]


def get_wind_char(wind: int, lang: int = 0) -> str:
    if 0 <= wind < 4:
        if lang == 1:
            return to_wind_char[wind + 4]
        return to_wind_char[wind]
    else:
        return " "


def get_modifier(from_who: Optional[RelativePlayerIdx], tile_unit_type: EventType) -> str:
    if from_who is None:
        return ""
    if tile_unit_type == EventType.ADDED_KAN:
        return to_relative_player_idx[from_who] + "(Add)"
    else:
        return to_relative_player_idx[from_who]


def get_yaku(yaku: int) -> str:
    return yaku_list[yaku]


def get_event_type(last_event: EventType, lang: int) -> str:
    if lang == 0:
        return event_type_en[last_event]
    else:
        return event_type_ja[last_event]



from __future__ import annotations  # postpone type hint evaluation or doctest fails

import argparse
import difflib
import os
import sys
from typing import List


def check_equality(original: str, restored: str) -> bool:
    original = original.strip().strip("\n")
    restored = restored.strip().strip("\n")
    assert len(original.split("\n")) == 1
    assert len(restored.split("\n")) == 1

    def split_mjlog(mjlog: str) -> List[str]:
        ret = []
        elem = ""
        for x in mjlog:
            elem += x
            if x == ">":
                ret.append(elem)
                elem = ""
        return ret

    for line in difflib.unified_diff(split_mjlog(original), split_mjlog(restored), n=0):
        if line.startswith("-<GO") or line.startswith("+<GO"):
            continue
        if line.startswith("-<INIT") or line.startswith("+<INIT"):
            continue
        if line.startswith("-<UN") or line.startswith("+<UN"):
            continue
        if line.startswith("-<SHUFFLE") or line.startswith("+<SHUFFLE"):
            continue
        if line.startswith("@@"):
            continue
        if line.startswith("-<BYE") or line.startswith("+<BYE"):
            continue
        if line.strip() in ["+", "---", "+++"]:
            continue
        if not line.strip():
            continue
        sys.stdout.write(line + "\n")
        return False
    return True


def load_mjlog(path: str) -> str:
    line = ""
    with open(path, "r") as f:
        lines = f.readlines()
        assert len(lines) == 1
        line = lines[0]
    return line


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="""Show diff between original mjlog and restored mjlog

    Example:

      $ python mjlog_decoder.py resources/mjlog resources/restored_mjlog
    """
    )
    parser.add_argument("original_dir", help="Path to original mjlogs")
    parser.add_argument("restored_dir", help="Path to restored mjlogs")
    args = parser.parse_args()

    original_paths = sorted(
        [
            os.path.join(args.original_dir, x)
            for x in os.listdir(args.original_dir)
            if x.endswith(".mjlog")
        ]
    )
    restored_paths = sorted(
        [
            os.path.join(args.restored_dir, x)
            for x in os.listdir(args.restored_dir)
            if x.endswith(".mjlog")
        ]
    )

    all_ok = True
    for original_path, restored_path in zip(original_paths, restored_paths):
        assert (
            original_path.split("/")[-1].split(".")[0]
            == restored_path.split("/")[-1].split(".")[0]
        )
        original = load_mjlog(original_path)
        restored = load_mjlog(restored_path)
        all_ok = all_ok and check_equality(original, restored)

    if not all_ok:
        exit(1)



import os
from typing import List

from mjx.converter import mjlog_decoder, mjlog_encoder
from mjx.utils.diff import check_equality


def test_encode_decode():
    decoder = mjlog_decoder.MjlogDecoder(modify=False)
    encoder = mjlog_encoder.MjlogEncoder()
    mjlog_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/mjlog")
    for mjlog in os.listdir(mjlog_dir):
        mjlog = os.path.join(mjlog_dir, mjlog)
        with open(mjlog, "r") as f:
            lines = f.readlines()
            assert len(lines) == 1
            original = lines[0]
            decoded: List[str] = decoder.decode(original)
            restored = encoder.encode(decoded)
            print(mjlog)
            assert check_equality(original, restored)



from typing import List

import mjxproto
from mjx.converter.open_converter import open_event_type
from mjx.converter.open_tile_ids_converter import open_tile_ids


class Hand:
    def __init__(self, closed_tiles: List[int] = [], opens: List[int] = []) -> None:
        self._closed_tiles: List[int] = closed_tiles  # each item is tile id (0 ~ 135)
        self.opens: List[int] = opens  # e.g., 49495. Follow Tenhou format.
        self.closed_tiles.sort()

    @property
    def closed_tiles(self) -> List[int]:
        self._closed_tiles.sort()
        return self._closed_tiles

    def add(self, tile_id: int) -> None:
        """Used for draw and ron"""
        assert len(self._closed_tiles) in [1, 4, 7, 10, 13]
        assert tile_id not in self._closed_tiles
        self.closed_tiles.append(tile_id)

    def discard(self, tile_id: int) -> None:
        assert len(self._closed_tiles) in [2, 5, 8, 11, 14]
        assert tile_id in self._closed_tiles
        self.closed_tiles.remove(tile_id)

    def apply_open(self, open: int) -> None:
        """
        Update closed_tiles and opens by open action.

        TODO: fix test cases

        >>> hand = Hand([2, 4, 6, 5, 11, 10, 8, 9, 86, 90, 125, 126, 130], [])
        >>> hand.apply_open(49495)  # chi
        >>> hand.closed_tiles
        [2, 4, 5, 6, 8, 9, 10, 11, 125, 126, 130]
        >>> hand.opens
        [49495]
        """
        if open_event_type(open) == mjxproto.EVENT_TYPE_ADDED_KAN:
            self._apply_added_kan(open)
            return

        ids = open_tile_ids(open)
        for id in ids:
            if id not in self.closed_tiles:
                continue
            self.closed_tiles.remove(id)

        self.opens.append(open)

        assert len(self.opens) == len(set(self.opens))

    def _apply_added_kan(self, open: int) -> None:
        assert open_event_type(open) == mjxproto.EVENT_TYPE_ADDED_KAN
        num_closed_tiles = len(self._closed_tiles)
        num_opens = len(self.opens)

        opens = self.opens
        for ix, old_open in enumerate(opens):
            if open_event_type(old_open) != mjxproto.EVENT_TYPE_PON:
                continue
            if open_tile_ids(old_open)[0] // 4 != open_tile_ids(open)[0] // 4:
                continue
            self.opens[ix] = open

        ids = open_tile_ids(open)
        for id in ids:
            if id not in self.closed_tiles:
                continue
            self.closed_tiles.remove(id)

        assert num_closed_tiles - 1 == len(self._closed_tiles)
        assert num_opens == len(self.opens)
        assert len(self.opens) == len(set(self.opens))



from __future__ import annotations  # postpone type hint evaluation or doctest fails

import json
import os
import sys
from typing import List, Optional

import click
from google.protobuf import json_format

import mjxproto
from mjx.converter.mjlog_decoder import MjlogDecoder
from mjx.converter.mjlog_encoder import MjlogEncoder
from mjx.visualizer.visualizer import GameBoardVisualizer, GameVisualConfig, MahjongTable


@click.group(help="A CLI tool of mjx")
def cli():
    pass


class LineBuffer:
    """Split lines of inputs by game end."""

    def __init__(self, fmt: str):
        self.fmt_: str = fmt
        self.curr_: List[str] = []
        self.buffer_: List[List[str]] = []

    @staticmethod
    def is_new_round_(line):
        d = json.loads(line)
        state = json_format.ParseDict(d, mjxproto.State())
        return (
            state.public_observation.init_score.round == 0
            and state.public_observation.init_score.honba == 0
        )

    def put(self, line: str) -> None:
        line = line.strip().strip("\n")
        if len(line) == 0:
            return
        if self.fmt_.startswith("mjxproto"):
            cnt = line.count("initScore")
            assert (
                cnt == 1
            ), f"Each line should only has one round but has {cnt}\nInput file may miss the last newline character."
            if LineBuffer.is_new_round_(line) and len(self.curr_) != 0:
                self.buffer_.append(self.curr_)
                self.curr_ = []
            self.curr_.append(line)
        elif self.fmt_ == "mjlog":
            cnt = line.count("</mjloggm>")
            assert (
                cnt == 1
            ), f"Each line should only has one game but has {cnt}\nInput file may miss the last newline character."
            self.buffer_.append([line])  # each line corresponds to each game

    def get(
        self, get_all: bool = False
    ) -> List[List[str]]:  # each List[str] corresponds to each game.
        if get_all and len(self.curr_) != 0:
            assert self.fmt_ != "mjlog"
            self.buffer_.append(self.curr_)
            self.curr_ = []
        tmp = self.buffer_
        self.buffer_ = []
        return tmp

    def empty(self) -> bool:
        return len(self.buffer_) == 0


def detect_format(line: str) -> str:
    try:
        json.loads(line)
        return "mjxproto"
    except ValueError:
        return "mjlog"


class Converter:
    def __init__(self, fmt_from: str, fmt_to: str, compress: bool):
        self.fmt_from: str = fmt_from
        self.fmt_to: str = fmt_to
        self.mjlog2mjxproto = None
        self.mjxproto2mjlog = None
        self.compress = compress

        if self.fmt_from == "mjxproto" and self.fmt_to == "mjlog":
            self.mjxproto2mjlog = MjlogEncoder()
        elif self.fmt_from == "mjlog" and self.fmt_to == "mjxproto":
            self.mjlog2mjxproto = MjlogDecoder(modify=True)
        elif self.fmt_from == "mjlog" and self.fmt_to == "mjxproto_raw":
            self.mjlog2mjxproto = MjlogDecoder(modify=False)
        else:
            sys.stderr.write(f"Input format = {self.fmt_from}\n")
            sys.stderr.write(f"Output format = {self.fmt_to}\n")
            raise ValueError("Input format and output format should be different")

    def convert(self, lines: List[str]) -> List[str]:
        """Transform a log of completed (or non-completed) one game into another format.

        :param lines: must correspond to completed (or non-completed) one game
        :return: also correspond to completed (or non-completed) one game
        """
        if self.fmt_from == "mjxproto" and self.fmt_to == "mjlog":
            assert self.mjxproto2mjlog is not None
            for line in lines:
                self.mjxproto2mjlog.put(line)
            return [self.mjxproto2mjlog.get()]  # a mjlog line corresponds to one game
        if self.fmt_from == "mjlog" and self.fmt_to == "mjxproto":
            assert self.mjlog2mjxproto is not None
            assert len(lines) == 1  # each line has each game
            return self.mjlog2mjxproto.decode(lines[0], compress=self.compress)
        if self.fmt_from == "mjlog" and self.fmt_to == "mjxproto_raw":
            assert self.mjlog2mjxproto is not None
            assert len(lines) == 1  # each line has each game
            return self.mjlog2mjxproto.decode(lines[0], compress=self.compress)
        else:
            raise NotImplementedError


class StdinIterator(object):
    def __iter__(self):
        try:
            line = sys.stdin.readline()
            while line:
                yield line
                line = sys.stdin.readline()
        except KeyboardInterrupt:
            return


@cli.command()
@click.argument("dir_from", type=str, default="")
@click.argument("dir_to", type=str, default="")
@click.option("--to-mjxproto", is_flag=True)
@click.option("--to-mjxproto-raw", is_flag=True)
@click.option("--to-mjlog", is_flag=True)
@click.option("--compress", is_flag=True)
@click.option("--verbose", is_flag=True)
def convert(
    dir_from: str,
    dir_to: str,
    to_mjxproto: bool,
    to_mjxproto_raw: bool,
    to_mjlog: bool,
    compress: bool,
    verbose: bool,
):
    """Convert Mahjong log into another format.

    Example (using stdin)

      $ cat test.mjlog | mjx convert --to-mjxproto
      $ cat test.mjlog | mjx convert --to-mjxproto-raw
      $ cat test.json  | mjx convert --to-mjlog

    Example (using file inputs)

    [NOTE] File inputs assume that each file corresponds to each game in any format.

      $ mjx convert ./mjlog_dir ./mjxproto_dir --to-mjxproto
      $ mjx convert ./mjlog_dir ./mjxproto_dir --to-mjxproto-raw
      $ mjx convert ./mjxproto_dir ./mjlog_dir --to-mjlog

    Difference between mjxproto and mjxproto-raw:

      1. Yaku is sorted in yaku number
      2. Yakuman's fu is set to 0
    """

    def to() -> str:
        assert (dir_from and dir_to) or (not dir_from and not dir_to)
        assert to_mjxproto or to_mjxproto_raw or to_mjlog

        if to_mjxproto:
            assert not (to_mjxproto_raw or to_mjlog)
            return "mjxproto"
        elif to_mjxproto_raw:
            assert not (to_mjxproto or to_mjlog)
            return "mjxproto_raw"
        elif to_mjlog:
            assert not (to_mjxproto or to_mjxproto_raw)
            return "mjlog"
        else:
            raise ValueError()

    fmt_from: str = ""
    converter: Optional[Converter] = None
    buffer: Optional[LineBuffer] = None

    if not dir_from and not dir_to:  # From stdin
        if verbose:
            sys.stderr.write(f"Converting to {to()}. stdin => stdout\n")

        itr = StdinIterator()
        for line in itr:
            line = line.strip().strip("\n")
            if len(line) == 0:
                continue

            if buffer is None or converter is None:
                assert buffer is None and converter is None
                fmt_from = detect_format(line)
                converter = Converter(fmt_from, to(), compress)
                buffer = LineBuffer(fmt_from)

            buffer.put(line)
            for lines in buffer.get():  # 終局時以外は空のはず
                for transformed_line in converter.convert(lines):
                    sys.stdout.write(transformed_line)

        # 終局で終わっていないときのため
        assert buffer is not None
        for lines in buffer.get(get_all=True):
            assert converter is not None
            for transformed_line in converter.convert(lines):
                sys.stdout.write(transformed_line)

    else:  # From files
        if verbose:
            sys.stderr.write(f"Converting to {to()}. {dir_from} => {dir_to}\n")

        to_type = to()
        to_ext = "mjlog" if to_type == "mjlog" else "json"
        num_mjlog = sum([1 for x in os.listdir(dir_from) if x.endswith("mjlog")])
        num_mjxproto = sum([1 for x in os.listdir(dir_from) if x.endswith("json")])
        assert not (
            num_mjlog > 0 and num_mjxproto > 0
        ), "There are two different formats in source directory."
        assert (
            num_mjlog > 0 or num_mjxproto > 0
        ), "There are no valid file formats in the source directory."
        for file_from in os.listdir(dir_from):
            if not file_from.endswith("json") and not file_from.endswith("mjlog"):
                continue

            path_from = os.path.join(dir_from, file_from)
            path_to = os.path.join(
                dir_to,
                os.path.splitext(os.path.basename(path_from))[0] + "." + to_ext,
            )

            if verbose:
                sys.stderr.write(f"Converting {path_from} to {path_to}\n")

            # 読み込み（全てのフォーマットで、１ファイル１半荘を想定）
            transformed_lines: List[str] = []
            with open(path_from, "r") as f:
                for line in f:
                    line = line.strip().strip("\n")
                    if len(line) == 0:
                        continue

                    if buffer is None or converter is None:
                        assert buffer is None and converter is None
                        fmt_from = detect_format(line)
                        converter = Converter(fmt_from, to(), compress)
                        buffer = LineBuffer(fmt_from)

                    buffer.put(line)

            # 変換
            assert buffer is not None
            list_lines: List[List[str]] = buffer.get(get_all=True)
            assert len(list_lines) == 1, "Each file should have one game"
            assert converter is not None
            transformed_lines += converter.convert(list_lines[0])

            # 書き込み
            with open(path_to, "w") as f:
                for line in transformed_lines:
                    f.write(line)


@cli.command()
@click.argument("path", type=str, default="")
@click.argument("page", type=str, default="0")
@click.option("--uni", is_flag=True)
@click.option("--rich", is_flag=True)
@click.option("--show_name", is_flag=True)
@click.option("--jp", is_flag=True)
def visualize(path: str, page: str, mode: str, uni: bool, rich: bool, show_name: bool, jp: bool):
    """Visualize Mahjong json data.

    Example (using stdin)

      $ cat test.json  | mjx visualize
      $ head test.json -n 10  | mjx visualize --rich --uni
      $ head test.json -n 10  | mjx visualize --rich --jp

    Example (using file inputs)

      $ mjx visualize test.json 0
      $ mjx visualize test.json 0 --rich --uni --show_name --jp

    """
    board_visualizer = GameBoardVisualizer(
        GameVisualConfig(rich=rich, uni=uni, show_name=show_name, lang=(1 if jp else 0))
    )

    if path == "":  # From stdin
        itr = StdinIterator()
        for line in itr:
            s_line = line.strip().strip("\n")
            proto_data = MahjongTable.json_to_proto(s_line)
            mahjong_table = MahjongTable.from_proto(proto_data)
            board_visualizer.print(mahjong_table)

    else:  # From files
        proto_data_list = MahjongTable.load_proto_data(path)
        mahjong_tables = [MahjongTable.from_proto(proto_data) for proto_data in proto_data_list]
        board_visualizer.print(mahjong_tables[int(page)])


def main():
    cli()


if __name__ == "__main__":
    main()



# postpone type hint evaluation or doctest fails
from __future__ import annotations

from typing import List

import mjxproto
from mjx.const import RelativePos


def open_event_type(bits: int) -> mjxproto.EventTypeValue:
    """
    >>> open_event_type(47723) == mjxproto.EVENT_TYPE_PON
    True
    >>> open_event_type(49495) == mjxproto.EVENT_TYPE_CHI
    True
    >>> open_event_type(28722) == mjxproto.EVENT_TYPE_ADDED_KAN
    True
    """
    if 1 << 2 & bits:
        return mjxproto.EVENT_TYPE_CHI
    elif 1 << 3 & bits:
        return mjxproto.EVENT_TYPE_PON
    elif 1 << 4 & bits:
        return mjxproto.EVENT_TYPE_ADDED_KAN
    else:
        if RelativePos.SELF == bits & 3:
            return mjxproto.EVENT_TYPE_CLOSED_KAN
        else:
            return mjxproto.EVENT_TYPE_OPEN_KAN


def open_from(bits: int) -> RelativePos:
    """
    >>> open_from(51306) == RelativePos.MID  # 対面
    True
    >>> open_from(49495) == RelativePos.LEFT  # 上家
    True
    >>> open_from(28722) == RelativePos.MID  # 加槓
    True
    """

    event_type = open_event_type(bits)
    if event_type == mjxproto.EVENT_TYPE_CHI:
        return RelativePos.LEFT
    elif (
        event_type == mjxproto.EVENT_TYPE_PON
        or event_type == mjxproto.EVENT_TYPE_OPEN_KAN
        or event_type == mjxproto.EVENT_TYPE_ADDED_KAN
    ):
        return RelativePos(bits & 3)
    else:
        return RelativePos.SELF


def _min_tile_chi(bits: int) -> int:
    x = (bits >> 10) // 3  # 0~21
    min_tile = (x // 7) * 9 + x % 7  # 0~33 base 9
    return min_tile


def is_stolen_red(bits: int, stolen_tile_kind) -> bool:  # TODO: test  さらに小さい関数を作るか否か考えるべし
    """
    >>> is_stolen_red(51306, 33)
    False
    >>> is_stolen_red(51247, 22)
    True
    """
    fives = [4, 13, 22]
    reds = [14, 52, 88]
    event_type = open_event_type(bits)
    if stolen_tile_kind not in fives:
        return False

    if event_type == mjxproto.EVENT_TYPE_CHI:
        stolen_tile_mod3 = (bits >> 10) % 3  # 鳴いた牌のindex
        stolen_tile_id_mod4 = (bits >> (3 + 2 * stolen_tile_mod3)) % 4  # 鳴いた牌のi
        return stolen_tile_id_mod4 == 0  # 鳴いた牌のid mod 4=0→赤
    elif event_type == mjxproto.EVENT_TYPE_PON or event_type == mjxproto.EVENT_TYPE_ADDED_KAN:
        unused_id_mod4 = (bits >> 5) % 4  # 未使用牌のid mod 4
        stolen_tile_mod3 = (bits >> 9) % 3  # 鳴いた牌のindex
        return unused_id_mod4 != 0 and stolen_tile_mod3 == 0  # 未使用牌が赤でなく、鳴いた牌のインデックスが0の時→赤
    else:
        return (bits >> 8) in reds


def is_unused_red(bits: int):
    unused_id_mod4 = (bits >> 5) % 4
    return unused_id_mod4 == 0


def has_red_chi(bits: int) -> bool:  # TODO テストgit
    min_starts_include5_mod9 = [2, 3, 4]
    min_tile = _min_tile_chi(bits)
    if min_tile % 9 not in min_starts_include5_mod9:
        return False
    else:
        start_from3 = min_tile % 9 == 2  # min_tile で場合分け
        start_from4 = min_tile % 9 == 3
        start_from5 = min_tile % 9 == 4
        if start_from3:  # 3から始まる→3番目の牌のid mod 4 =0 →赤
            return (bits >> 7) % 4 == 0
        elif start_from4:
            return (bits >> 5) % 4 == 0
        elif start_from5:
            return (bits >> 3) % 4 == 0
        else:
            assert False


# TODO テスト ポンとカカンは未使用牌が赤かどうかで鳴牌に赤があるか判断
def has_red_pon_kan_added(bits: int) -> bool:
    fives = [4, 13, 22, 51, 52, 53]
    stolen_tile_kind = open_stolen_tile_type(bits)
    if stolen_tile_kind in fives:
        unused_id_mod4 = (bits >> 5) % 4
        return unused_id_mod4 != 0
    else:
        return False


def has_red_kan_closed_kan_opend(bits: int) -> bool:
    fives = [4, 13, 22, 51, 52, 53]
    stolen_tile_kind = open_stolen_tile_type(bits)
    return stolen_tile_kind in fives


def has_red(bits: int) -> bool:
    """
    >>> has_red(52503)  # 赤５を含むチー
    True
    """
    event_type = open_event_type(bits)
    if event_type == mjxproto.EVENT_TYPE_CHI:
        return has_red_chi(bits)
    elif event_type == mjxproto.EVENT_TYPE_PON or event_type == mjxproto.EVENT_TYPE_ADDED_KAN:
        return has_red_pon_kan_added(bits)
    else:
        return has_red_kan_closed_kan_opend(bits)  # ダイミンカンとアンカンは必ず赤を含む


def transform_red_stolen(bits: int, stolen_tile: int) -> int:
    red_dict = {4: 51, 13: 52, 22: 53}  # openの5:mjscoreの赤５
    if is_stolen_red(bits, stolen_tile):
        return red_dict[stolen_tile]
    else:
        return stolen_tile


def transform_red_open(bits: int, open: List[int], event_type) -> List[int]:
    """
    >>> transform_red_open(52503, [21, 22, 23], mjxproto.EVENT_TYPE_CHI)
    [21, 53, 23]
    """
    red_dict = {4: 51, 13: 52, 22: 53}
    fives = [4, 13, 22]
    if not has_red(bits):
        return open
    if event_type == mjxproto.EVENT_TYPE_CHI:
        return [red_dict[i] if i in fives else i for i in open]
    else:
        open[-1] = red_dict[open[-1]]
        return open


def open_stolen_tile_type(bits: int) -> int:
    """
    >>> open_stolen_tile_type(51306)
    33
    >>> open_stolen_tile_type(49495)
    20
    >>> open_stolen_tile_type(28722)
    18
    >>> open_stolen_tile_type(31744)  #  暗槓　白
    31
    """
    event_type = open_event_type(bits)
    if event_type == mjxproto.EVENT_TYPE_CHI:
        min_tile = _min_tile_chi(bits)
        stolen_tile_kind = min_tile + (bits >> 10) % 3
        return transform_red_stolen(bits, stolen_tile_kind)
    elif event_type == mjxproto.EVENT_TYPE_PON or event_type == mjxproto.EVENT_TYPE_ADDED_KAN:
        stolen_tile_kind = (bits >> 9) // 3
        return transform_red_stolen(bits, stolen_tile_kind)
    else:
        stolen_tile_kind = (bits >> 8) // 4
        return transform_red_stolen(bits, stolen_tile_kind)


def open_tile_types(bits: int) -> List[int]:
    """
    >>> open_tile_types(51306)  # Pon rd
    [33, 33, 33]
    >>> open_tile_types(49495)  # Chi s3s4s5
    [20, 21, 22]
    >>> open_tile_types(28722)  # 加槓 s1
    [18, 18, 18, 18]
    >>> open_tile_types(31744)  # 暗槓　白
    [31, 31, 31, 31]
    """
    reds = [51, 52, 53]
    red_five_dict = {51: 4, 52: 13, 53: 22}
    event_type = open_event_type(bits)
    if event_type == mjxproto.EVENT_TYPE_CHI:
        min_tile = _min_tile_chi(bits)
        open = [min_tile, min_tile + 1, min_tile + 2]
        return transform_red_open(bits, open, event_type)
    stolen_tile_kind = open_stolen_tile_type(bits)
    if stolen_tile_kind in reds:  # 赤だった場合はopenのformatに置換しないと、複数の同じ赤を持ったopenが生成される。
        stolen_tile_kind = red_five_dict[stolen_tile_kind]
    else:
        None
    if event_type == mjxproto.EVENT_TYPE_PON:
        open = [stolen_tile_kind] * 3
        return transform_red_open(bits, open, event_type)
    else:
        open = [stolen_tile_kind] * 4
        return transform_red_open(bits, open, event_type)


def change_open_tile_fmt(
    tile_in_open_fmt: int,
) -> int:  # tile_in_open 0~33 tile_in_score 11~19, 21~29, 31~39,41~47
    reds_in_score = [51, 52, 53]
    if tile_in_open_fmt in reds_in_score:
        return tile_in_open_fmt
    else:
        tile_in_score = 10 + 10 * (tile_in_open_fmt // 9) + (tile_in_open_fmt % 9 + 1)
        return tile_in_score


def change_open_tiles_fmt(tile_ids_in_open: List[int]) -> List[int]:
    """
    >>> change_open_tiles_fmt([21, 22, 23])
    [34, 35, 36]
    """
    scores = list(map(change_open_tile_fmt, tile_ids_in_open))
    return scores



from __future__ import annotations

from typing import List, Optional

import _mjx  # type: ignore
from google.protobuf import json_format

import mjxproto
from mjx.const import TileType
from mjx.open import Open
from mjx.tile import Tile


class Hand:
    def __init__(self, hand_json: Optional[str] = None) -> None:
        self._cpp_obj: Optional[_mjx.Hand] = None  # type: ignore
        if hand_json is None:
            return

        self._cpp_obj = _mjx.Hand(hand_json)  # type: ignore

    def closed_tiles(self) -> List[Tile]:
        assert self._cpp_obj is not None  # type: ignore
        return [Tile(tid) for tid in self._cpp_obj.closed_tiles()]  # type: ignore

    def closed_tile_types(self) -> List[TileType]:
        assert self._cpp_obj is not None  # type: ignore
        return [TileType(tt) for tt in self._cpp_obj.closed_tile_types()]  # type: ignore

    def opens(self) -> List[Open]:
        assert self._cpp_obj is not None  # type: ignore
        return [Open(bit) for bit in self._cpp_obj.opens()]  # type: ignore

    def shanten_number(self) -> int:
        assert self._cpp_obj is not None  # type: ignore
        return self._cpp_obj.shanten_number()  # type: ignore

    def effective_draw_types(self) -> List[TileType]:
        return [TileType(i) for i in self._cpp_obj.effective_draw_types()]  # type: ignore

    def effective_discard_types(self) -> List[TileType]:
        return [TileType(i) for i in self._cpp_obj.effective_discard_types()]  # type: ignore

    def to_json(self) -> str:
        assert self._cpp_obj is not None  # type: ignore
        return self._cpp_obj.to_json()  # type: ignore

    def to_proto(self) -> mjxproto.Hand:
        assert self._cpp_obj is not None  # type: ignore
        return json_format.Parse(self.to_json(), mjxproto.Hand())

    @classmethod
    def _from_cpp_obj(cls, cpp_obj: _mjx.Hand) -> Hand:  # type: ignore
        hand = cls()
        hand._cpp_obj = cpp_obj  # type: ignore
        return hand
